# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
import os
import re
import stat
import tempfile
from textwrap import dedent
from unittest.mock import Mock, patch

import fixtures
import pytest
from testtools.matchers import Contains, Equals, FileExists, Not

import snapcraft
from snapcraft.internal import (
    common,
    errors,
    lifecycle,
    pluginhandler,
    project_loader,
    repo,
    states,
    steps,
)
from snapcraft.internal.sources.errors import SnapcraftSourceUnhandledError
from snapcraft.project import Project
from tests import fixture_setup, unit

from . import mocks


class PluginTestCase(unit.TestCase):
    def test_build_with_subdir_copies_sourcedir(self):
        handler = self.load_part("test-part", part_properties={"source-subdir": "src"})

        sourcedir = handler.part_source_dir
        source_subdir = handler.plugin.options.source_subdir

        subdir = os.path.join(sourcedir, source_subdir)
        os.makedirs(subdir)
        open(os.path.join(sourcedir, "file1"), "w").close()
        open(os.path.join(subdir, "file2"), "w").close()

        self.assertThat(
            handler.part_build_work_dir,
            Equals(os.path.join(handler.plugin.build_basedir, source_subdir)),
        )

        handler.build()

        self.assertThat(os.path.join(handler.part_build_dir, "file1"), FileExists())
        self.assertThat(
            os.path.join(handler.part_build_dir, source_subdir, "file2"), FileExists()
        )

    def test_build_with_missing_metadata_file(self):
        handler = self.load_part(
            "test-part", part_properties={"parse-info": ["missing-file"]}
        )
        handler.makedirs()

        raised = self.assertRaises(errors.MissingMetadataFileError, handler.build)
        self.assertThat(raised.path, Equals("missing-file"))

    def test_build_without_subdir_copies_sourcedir(self):
        handler = self.load_part("test-part")

        os.makedirs(handler.part_source_dir)
        open(os.path.join(handler.part_source_dir, "file"), "w").close()

        self.assertThat(handler.part_build_dir, Equals(handler.plugin.build_basedir))

        handler.build()

        self.assertTrue(
            os.path.exists(os.path.join(handler.plugin.build_basedir, "file"))
        )

    @patch("os.path.isdir", return_value=False)
    def test_local_non_dir_source_path_must_raise_exception(self, mock_isdir):
        self.assertRaises(
            SnapcraftSourceUnhandledError,
            self.load_part,
            "test-part",
            part_properties={"source": "file"},
        )

        mock_isdir.assert_any_call("file")

    def test_fileset_include_excludes(self):
        stage_set = [
            "-etc",
            "opt/something",
            "-usr/lib/*.a",
            "usr/bin",
            r"\-everything",
            r"\\a",
        ]

        include, exclude = pluginhandler._get_file_list(stage_set)

        self.assertThat(
            include, Equals(["opt/something", "usr/bin", "-everything", r"\a"])
        )
        self.assertThat(exclude, Equals(["etc", "usr/lib/*.a"]))

    @patch.object(snapcraft.plugins.v1.nil.NilPlugin, "snap_fileset")
    def test_migratable_fileset_for_no_options_modification(self, mock_snap_fileset):
        """Making sure migratable_fileset_for() doesn't modify options"""

        mock_snap_fileset.return_value = ["baz"]

        handler = self.load_part("test-part")
        handler.plugin.options.snap = ["foo"]
        handler.plugin.options.stage = ["bar"]
        expected_options = copy.deepcopy(handler.plugin.options)

        handler.migratable_fileset_for(steps.STAGE)
        self.assertThat(
            vars(handler.plugin.options),
            Equals(vars(expected_options)),
            "Expected options to be unmodified",
        )

        handler.migratable_fileset_for(steps.BUILD)
        self.assertThat(
            vars(handler.plugin.options),
            Equals(vars(expected_options)),
            "Expected options to be unmodified",
        )

    def test_fileset_only_includes(self):
        stage_set = ["opt/something", "usr/bin"]

        include, exclude = pluginhandler._get_file_list(stage_set)

        self.assertThat(include, Equals(["opt/something", "usr/bin"]))
        self.assertThat(exclude, Equals([]))

    def test_fileset_only_excludes(self):
        stage_set = ["-etc", "-usr/lib/*.a"]

        include, exclude = pluginhandler._get_file_list(stage_set)

        self.assertThat(include, Equals(["*"]))
        self.assertThat(exclude, Equals(["etc", "usr/lib/*.a"]))

    def test_migrate_snap_files_already_exists(self):
        os.makedirs("install")
        os.makedirs("stage")

        # Place the already-staged file
        with open("stage/foo", "w") as f:
            f.write("staged")

        # Place the to-be-staged file with the same name
        with open("install/foo", "w") as f:
            f.write("installed")

        files, dirs = pluginhandler._migratable_filesets(["*"], "install")
        pluginhandler._migrate_files(files, dirs, "install", "stage")

        # Verify that the staged file is the one that was staged last
        with open("stage/foo", "r") as f:
            self.assertThat(
                f.read(),
                Equals("installed"),
                "Expected staging to allow overwriting of already-staged files",
            )

    def test_migrate_files_supports_no_follow_symlinks(self):
        os.makedirs("install")
        os.makedirs("stage")

        with open(os.path.join("install", "foo"), "w") as f:
            f.write("installed")

        os.symlink("foo", os.path.join("install", "bar"))

        files, dirs = pluginhandler._migratable_filesets(["*"], "install")
        pluginhandler._migrate_files(
            files, dirs, "install", "stage", follow_symlinks=False
        )

        # Verify that the symlink was preserved
        self.assertTrue(
            os.path.islink(os.path.join("stage", "bar")),
            "Expected migrated 'bar' to still be a symlink.",
        )
        self.assertThat(
            os.readlink(os.path.join("stage", "bar")),
            Equals("foo"),
            "Expected migrated 'bar' to point to 'foo'",
        )

    def test_migrate_files_preserves_symlink_file(self):
        os.makedirs("install")
        os.makedirs("stage")

        with open(os.path.join("install", "foo"), "w") as f:
            f.write("installed")

        os.symlink("foo", os.path.join("install", "bar"))

        files, dirs = pluginhandler._migratable_filesets(["*"], "install")
        pluginhandler._migrate_files(files, dirs, "install", "stage")

        # Verify that the symlinks were preserved
        self.assertTrue(
            os.path.islink(os.path.join("stage", "bar")),
            "Expected migrated 'sym-a' to be a symlink.",
        )

    def test_migrate_files_no_follow_symlinks(self):
        os.makedirs("install/usr/bin")
        os.makedirs("stage")

        with open(os.path.join("install", "usr", "bin", "foo"), "w") as f:
            f.write("installed")

        os.symlink("usr/bin", os.path.join("install", "bin"))

        files, dirs = pluginhandler._migratable_filesets(["-usr"], "install")
        pluginhandler._migrate_files(files, dirs, "install", "stage")

        # Verify that the symlinks were preserved
        assert files == {"bin"}
        assert dirs == set()

        self.assertTrue(
            os.path.islink(os.path.join("stage", "bin")),
            "Expected migrated 'bin' to be a symlink.",
        )

    def test_migrate_files_preserves_symlink_nested_file(self):
        os.makedirs(os.path.join("install", "a"))
        os.makedirs("stage")

        with open(os.path.join("install", "a", "foo"), "w") as f:
            f.write("installed")

        os.symlink(os.path.join("a", "foo"), os.path.join("install", "bar"))
        os.symlink(os.path.join("foo"), os.path.join("install", "a", "bar"))

        files, dirs = pluginhandler._migratable_filesets(["*"], "install")
        pluginhandler._migrate_files(files, dirs, "install", "stage")

        # Verify that the symlinks were preserved
        self.assertTrue(
            os.path.islink(os.path.join("stage", "bar")),
            "Expected migrated 'sym-a' to be a symlink.",
        )

        self.assertTrue(
            os.path.islink(os.path.join("stage", "a", "bar")),
            "Expected migrated 'sym-a' to be a symlink.",
        )

    def test_migrate_files_preserves_symlink_empty_dir(self):
        os.makedirs(os.path.join("install", "foo"))
        os.makedirs("stage")

        os.symlink("foo", os.path.join("install", "bar"))

        files, dirs = pluginhandler._migratable_filesets(["*"], "install")
        pluginhandler._migrate_files(files, dirs, "install", "stage")

        # Verify that the symlinks were preserved
        self.assertTrue(
            os.path.islink(os.path.join("stage", "bar")),
            "Expected migrated 'sym-a' to be a symlink.",
        )

    def test_migrate_files_preserves_symlink_nonempty_dir(self):
        os.makedirs(os.path.join("install", "foo"))
        os.makedirs("stage")

        os.symlink("foo", os.path.join("install", "bar"))

        with open(os.path.join("install", "foo", "xfile"), "w") as f:
            f.write("installed")

        files, dirs = pluginhandler._migratable_filesets(["*"], "install")
        pluginhandler._migrate_files(files, dirs, "install", "stage")

        # Verify that the symlinks were preserved
        self.assertTrue(
            os.path.islink(os.path.join("stage", "bar")),
            "Expected migrated 'sym-a' to be a symlink.",
        )

    def test_migrate_files_preserves_symlink_nested_dir(self):
        os.makedirs(os.path.join("install", "a", "b"))
        os.makedirs("stage")

        os.symlink(os.path.join("a", "b"), os.path.join("install", "bar"))
        os.symlink(os.path.join("b"), os.path.join("install", "a", "bar"))

        with open(os.path.join("install", "a", "b", "xfile"), "w") as f:
            f.write("installed")

        files, dirs = pluginhandler._migratable_filesets(["*"], "install")
        pluginhandler._migrate_files(files, dirs, "install", "stage")

        # Verify that the symlinks were preserved
        self.assertTrue(
            os.path.islink(os.path.join("stage", "bar")),
            "Expected migrated 'bar' to be a symlink.",
        )

        self.assertTrue(
            os.path.islink(os.path.join("stage", "a", "bar")),
            "Expected migrated 'a/bar' to be a symlink.",
        )

    def test_migrate_files_supports_follow_symlinks(self):
        os.makedirs("install")
        os.makedirs("stage")

        with open(os.path.join("install", "foo"), "w") as f:
            f.write("installed")

        os.symlink("foo", os.path.join("install", "bar"))

        files, dirs = pluginhandler._migratable_filesets(["*"], "install")
        pluginhandler._migrate_files(
            files, dirs, "install", "stage", follow_symlinks=True
        )

        # Verify that the symlink was preserved
        self.assertFalse(
            os.path.islink(os.path.join("stage", "bar")),
            "Expected migrated 'bar' to no longer be a symlink.",
        )
        with open(os.path.join("stage", "bar"), "r") as f:
            self.assertThat(
                f.read(),
                Equals("installed"),
                "Expected migrated 'bar' to be a copy of 'foo'",
            )

    def test_migrate_files_preserves_file_mode(self):
        os.makedirs("install")
        os.makedirs("stage")

        foo = os.path.join("install", "foo")

        with open(foo, "w") as f:
            f.write("installed")

        mode = os.stat(foo).st_mode

        new_mode = 0o777
        os.chmod(foo, new_mode)
        self.assertNotEqual(mode, new_mode)

        files, dirs = pluginhandler._migratable_filesets(["*"], "install")
        pluginhandler._migrate_files(
            files, dirs, "install", "stage", follow_symlinks=True
        )

        self.assertThat(
            new_mode,
            Equals(stat.S_IMODE(os.stat(os.path.join("stage", "foo")).st_mode)),
        )

    def test_migrate_files_preserves_file_mode_chown_permissions(self):
        os.makedirs("install")
        os.makedirs("stage")

        foo = os.path.join("install", "foo")

        with open(foo, "w") as f:
            f.write("installed")

        mode = os.stat(foo).st_mode

        new_mode = 0o777
        os.chmod(foo, new_mode)
        self.assertNotEqual(mode, new_mode)

        files, dirs = pluginhandler._migratable_filesets(["*"], "install")
        pluginhandler._migrate_files(
            files, dirs, "install", "stage", follow_symlinks=True
        )

        self.assertThat(
            new_mode,
            Equals(stat.S_IMODE(os.stat(os.path.join("stage", "foo")).st_mode)),
        )

    def test_migrate_files_preserves_directory_mode(self):
        os.makedirs("install/foo")
        os.makedirs("stage")

        foo = os.path.join("install", "foo", "bar")

        with open(foo, "w") as f:
            f.write("installed")

        mode = os.stat(foo).st_mode

        new_mode = 0o777
        self.assertNotEqual(mode, new_mode)
        os.chmod(os.path.dirname(foo), new_mode)
        os.chmod(foo, new_mode)

        files, dirs = pluginhandler._migratable_filesets(["*"], "install")
        pluginhandler._migrate_files(
            files, dirs, "install", "stage", follow_symlinks=True
        )

        self.assertThat(
            new_mode,
            Equals(stat.S_IMODE(os.stat(os.path.join("stage", "foo")).st_mode)),
        )
        self.assertThat(
            new_mode,
            Equals(stat.S_IMODE(os.stat(os.path.join("stage", "foo", "bar")).st_mode)),
        )

    def test_filesets_includes_without_relative_paths(self):
        raised = self.assertRaises(
            errors.PluginError, pluginhandler._get_file_list, ["rel", "/abs/include"]
        )

        self.assertThat(raised.message, Equals('path "/abs/include" must be relative'))

    def test_filesets_excludes_without_relative_paths(self):
        raised = self.assertRaises(
            errors.PluginError, pluginhandler._get_file_list, ["rel", "-/abs/exclude"]
        )

        self.assertThat(raised.message, Equals('path "/abs/exclude" must be relative'))

    @patch("snapcraft.internal.pluginhandler._organize_filesets")
    def test_build_organizes(self, mock_organize):
        handler = self.load_part("test-part")
        handler.build()
        mock_organize.assert_called_once_with(
            "test-part", {}, handler.part_install_dir, False
        )

    @patch("snapcraft.internal.pluginhandler._organize_filesets")
    def test_update_build_organizes_with_overwrite(self, mock_organize):
        class TestPlugin(snapcraft.BasePlugin):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.out_of_source_build = True

        self.useFixture(fixture_setup.FakePlugin("test-plugin", TestPlugin))

        handler = self.load_part("test-part", plugin_name="test-plugin")
        handler.makedirs()
        handler.update_build()
        mock_organize.assert_called_once_with(
            "test-part", {}, handler.part_install_dir, True
        )


class TestMigratePartFiles:

    scenarios = [
        ("nothing", {"fileset": ["-*"], "result": []}),
        (
            "all",
            {
                "fileset": ["*"],
                "result": [
                    "stage/1",
                    "stage/1/1a/1b",
                    "stage/1/1a",
                    "stage/1/a",
                    "stage/2",
                    "stage/2/2a",
                    "stage/3",
                    "stage/3/a",
                    "stage/a",
                    "stage/b",
                ],
            },
        ),
        (
            "no1",
            {
                "fileset": ["-1"],
                "result": [
                    "stage/2",
                    "stage/2/2a",
                    "stage/3",
                    "stage/3/a",
                    "stage/a",
                    "stage/b",
                ],
            },
        ),
        ("onlya", {"fileset": ["a"], "result": ["stage/a"]}),
        (
            "onlybase",
            {
                "fileset": ["*", "-*/*"],
                "result": ["stage/a", "stage/b", "stage/1", "stage/2", "stage/3"],
            },
        ),
        (
            "nostara",
            {
                "fileset": ["-*/a"],
                "result": [
                    "stage/1",
                    "stage/1/1a/1b",
                    "stage/1/1a",
                    "stage/2",
                    "stage/2/2a",
                    "stage/3",
                    "stage/a",
                    "stage/b",
                ],
            },
        ),
    ]

    def test_migrate_snap_files(self, tmp_path, fileset, result):
        srcdir = tmp_path / "install"

        (srcdir / "1/1a/1b").mkdir(parents=True)
        (srcdir / "2/2a").mkdir(parents=True)
        (srcdir / "3").mkdir(parents=True)

        (srcdir / "a").touch()
        (srcdir / "b").touch()
        (srcdir / "1/a").touch()
        (srcdir / "3/a").touch()

        dstdir = tmp_path / "stage"
        dstdir.mkdir(parents=True)

        files, dirs = pluginhandler._migratable_filesets(fileset, srcdir.as_posix())
        pluginhandler._migrate_files(files, dirs, srcdir.as_posix(), dstdir.as_posix())

        expected = []
        for item in result:
            expected.append(os.path.join(tmp_path, item))
        expected.sort()

        result = []
        for root, subdirs, files in os.walk(dstdir):
            for item in files:
                result.append(os.path.join(root, item))
            for item in subdirs:
                result.append(os.path.join(root, item))
        result.sort()

        assert result == expected


class MigratableFilesetsTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        os.makedirs("install/foo/bar/baz")
        open("install/1", "w").close()
        open("install/foo/2", "w").close()
        open("install/foo/bar/3", "w").close()
        open("install/foo/bar/baz/4", "w").close()

    def test_migratable_filesets_everything(self):
        files, dirs = pluginhandler._migratable_filesets(["*"], "install")
        self.assertThat(files, Equals({"1", "foo/2", "foo/bar/3", "foo/bar/baz/4"}))
        self.assertThat(dirs, Equals({"foo", "foo/bar", "foo/bar/baz"}))

    def test_migratable_filesets_foo(self):
        files, dirs = pluginhandler._migratable_filesets(["foo"], "install")
        self.assertThat(files, Equals({"foo/2", "foo/bar/3", "foo/bar/baz/4"}))
        self.assertThat(dirs, Equals({"foo", "foo/bar", "foo/bar/baz"}))

    def test_migratable_filesets_everything_in_foo(self):
        files, dirs = pluginhandler._migratable_filesets(["foo/*"], "install")
        self.assertThat(files, Equals({"foo/2", "foo/bar/3", "foo/bar/baz/4"}))
        self.assertThat(dirs, Equals({"foo", "foo/bar", "foo/bar/baz"}))

    def test_migratable_filesets_root_file(self):
        files, dirs = pluginhandler._migratable_filesets(["1"], "install")
        self.assertThat(files, Equals({"1"}))
        self.assertThat(dirs, Equals(set()))

    def test_migratable_filesets_single_nested_file(self):
        files, dirs = pluginhandler._migratable_filesets(["foo/2"], "install")
        self.assertThat(files, Equals({"foo/2"}))
        self.assertThat(dirs, Equals({"foo"}))

    def test_migratable_filesets_single_really_nested_file(self):
        files, dirs = pluginhandler._migratable_filesets(["foo/bar/2"], "install")
        self.assertThat(files, Equals({"foo/bar/2"}))
        self.assertThat(dirs, Equals({"foo", "foo/bar"}))

    def test_migratable_filesets_single_really_really_nested_file(self):
        files, dirs = pluginhandler._migratable_filesets(["foo/bar/baz/3"], "install")
        self.assertThat(files, Equals({"foo/bar/baz/3"}))
        self.assertThat(dirs, Equals({"foo", "foo/bar", "foo/bar/baz"}))


class TestOrganize:

    scenarios = [
        (
            "simple_file",
            dict(
                setup_dirs=[],
                setup_files=["foo"],
                organize_set={"foo": "bar"},
                expected=[(["bar"], "")],
                expected_message=None,
                expected_overwrite=None,
            ),
        ),
        (
            "simple_dir_with_file",
            dict(
                setup_dirs=["foodir"],
                setup_files=[os.path.join("foodir", "foo")],
                organize_set={"foodir": "bardir"},
                expected=[(["bardir"], ""), (["foo"], "bardir")],
                expected_message=None,
                expected_overwrite=None,
            ),
        ),
        (
            "organize_to_the_same_directory",
            dict(
                setup_dirs=["bardir", "foodir"],
                setup_files=[
                    os.path.join("foodir", "foo"),
                    os.path.join("bardir", "bar"),
                    "basefoo",
                ],
                organize_set={
                    "foodir": "bin",
                    "bardir": "bin",
                    "basefoo": "bin/basefoo",
                },
                expected=[(["bin"], ""), (["bar", "basefoo", "foo"], "bin")],
                expected_message=None,
                expected_overwrite=None,
            ),
        ),
        (
            "leading_slash_in_value",
            dict(
                setup_dirs=[],
                setup_files=["foo"],
                organize_set={"foo": "/bar"},
                expected=[(["bar"], "")],
                expected_message=None,
                expected_overwrite=None,
            ),
        ),
        (
            "overwrite_existing_file",
            dict(
                setup_dirs=[],
                setup_files=["foo", "bar"],
                organize_set={"foo": "bar"},
                expected=errors.SnapcraftOrganizeError,
                expected_message=r".*trying to organize file 'foo' to 'bar', but 'bar' already exists.*",
                expected_overwrite=[(["bar"], "")],
            ),
        ),
        (
            "*_for_files",
            dict(
                setup_dirs=[],
                setup_files=["foo.conf", "bar.conf"],
                organize_set={"*.conf": "dir/"},
                expected=[(["dir"], ""), (["bar.conf", "foo.conf"], "dir")],
                expected_message=None,
                expected_overwrite=None,
            ),
        ),
        (
            "*_for_files_with_non_dir_dst",
            dict(
                setup_dirs=[],
                setup_files=["foo.conf", "bar.conf"],
                organize_set={"*.conf": "dir"},
                expected=errors.SnapcraftOrganizeError,
                expected_message=r".*multiple files to be organized into 'dir'.*",
                expected_overwrite=None,
            ),
        ),
        (
            "*_for_directories",
            dict(
                setup_dirs=["dir1", "dir2"],
                setup_files=[os.path.join("dir1", "foo"), os.path.join("dir2", "bar")],
                organize_set={"dir*": "dir/"},
                expected=[
                    (["dir"], ""),
                    (["dir1", "dir2"], "dir"),
                    (["foo"], os.path.join("dir", "dir1")),
                    (["bar"], os.path.join("dir", "dir2")),
                ],
                expected_message=None,
                expected_overwrite=None,
            ),
        ),
        (
            "combined_*_with_file",
            dict(
                setup_dirs=["dir1", "dir2"],
                setup_files=[
                    os.path.join("dir1", "foo"),
                    os.path.join("dir1", "bar"),
                    os.path.join("dir2", "bar"),
                ],
                organize_set={"dir*": "dir/", "dir1/bar": "."},
                expected=[
                    (["bar", "dir"], ""),
                    (["dir1", "dir2"], "dir"),
                    (["foo"], os.path.join("dir", "dir1")),
                    (["bar"], os.path.join("dir", "dir2")),
                ],
                expected_message=None,
                expected_overwrite=None,
            ),
        ),
        (
            "*_into_dir",
            dict(
                setup_dirs=["dir"],
                setup_files=[os.path.join("dir", "foo"), os.path.join("dir", "bar")],
                organize_set={"dir/f*": "nested/dir/"},
                expected=[
                    (["dir", "nested"], ""),
                    (["bar"], "dir"),
                    (["dir"], "nested"),
                    (["foo"], os.path.join("nested", "dir")),
                ],
                expected_message=None,
                expected_overwrite=None,
            ),
        ),
    ]

    def _organize_and_assert(
        self,
        tmp_path,
        setup_dirs,
        setup_files,
        organize_set,
        expected,
        expected_message,
        expected_overwrite,
        overwrite,
    ):
        base_dir = tmp_path / "install"
        base_dir.mkdir(parents=True, exist_ok=True)

        for directory in setup_dirs:
            (base_dir / directory).mkdir(exist_ok=True)

        for file_entry in setup_files:
            (base_dir / file_entry).touch()

        if overwrite and expected_overwrite is not None:
            expected = expected_overwrite

        if isinstance(expected, type) and issubclass(expected, Exception):
            with pytest.raises(expected) as error:
                pluginhandler._organize_filesets(
                    "part-name", organize_set, base_dir, overwrite
                )
            assert re.match(expected_message, str(error)) is not None

        else:
            pluginhandler._organize_filesets(
                "part-name", organize_set, base_dir, overwrite
            )
            for expect in expected:
                dir_path = (base_dir / expect[1]).as_posix()
                dir_contents = os.listdir(dir_path)
                dir_contents.sort()
                assert dir_contents == expect[0]

    def test_organize(
        self,
        tmp_path,
        setup_dirs,
        setup_files,
        organize_set,
        expected,
        expected_message,
        expected_overwrite,
    ):
        self._organize_and_assert(
            tmp_path,
            setup_dirs,
            setup_files,
            organize_set,
            expected,
            expected_message,
            expected_overwrite,
            False,
        )

        # Verify that it can be organized again by overwriting
        self._organize_and_assert(
            tmp_path,
            setup_dirs,
            setup_files,
            organize_set,
            expected,
            expected_message,
            expected_overwrite,
            True,
        )


class RealStageTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        fake_install_build_packages = fixtures.MockPatch(
            "snapcraft.internal.lifecycle._runner._install_build_packages",
            return_value=list(),
        )
        self.useFixture(fake_install_build_packages)

        fake_install_build_snaps = fixtures.MockPatch(
            "snapcraft.internal.lifecycle._runner._install_build_snaps",
            return_value=list(),
        )
        self.useFixture(fake_install_build_snaps)

    def make_snapcraft_project(self):
        snapcraft_yaml_file_path = self.make_snapcraft_yaml(
            dedent(
                """\
            name: pc-file-test
            base: core18
            version: "1.0"
            summary: test pkg-config .pc
            description: when the .pc files reach stage the should be reprefixed
            confinement: strict
            grade: stable

            parts:
                stage-pc:
                    plugin: nil
        """
            )
        )
        project = Project(snapcraft_yaml_file_path=snapcraft_yaml_file_path)
        return project_loader.load_config(project)

    def test_pc_files_correctly_prefixed(self):
        pc_file = os.path.join("usr", "lib", "pkgconfig", "granite.pc")
        stage_pc_install = os.path.join(self.parts_dir, "stage-pc", "install", pc_file)
        stage_pc_stage = os.path.join(self.stage_dir, pc_file)

        # Run build
        project_config = self.make_snapcraft_project()
        lifecycle.execute(steps.BUILD, project_config)

        # Simulate a .pc file was installed
        os.makedirs(os.path.dirname(stage_pc_install))
        with open(stage_pc_install, "w") as f:
            f.write("prefix=/usr\n")
            f.write("exec_prefix=${prefix}\n")
            f.write("libdir=${prefix}/lib\n")
            f.write("includedir=${prefix}/include\n")
            f.write("\n")
            f.write("Name: granite\n")
            f.write("Description: elementary's Application Framework\n")
            f.write("Version: 0.4\n")
            f.write("Libs: -L${libdir} -lgranite\n")
            f.write("Cflags: -I${includedir}/granite\n")
            f.write("Requires: cairo gee-0.8 glib-2.0 gio-unix-2.0 gobject-2.0\n")

        # Now we stage
        lifecycle.execute(steps.STAGE, project_config)

        with open(stage_pc_stage) as f:
            pc_file_content = f.read()
        expected_pc_file_content = """prefix={}/usr
exec_prefix=${{prefix}}
libdir=${{prefix}}/lib
includedir=${{prefix}}/include

Name: granite
Description: elementary's Application Framework
Version: 0.4
Libs: -L${{libdir}} -lgranite
Cflags: -I${{includedir}}/granite
Requires: cairo gee-0.8 glib-2.0 gio-unix-2.0 gobject-2.0
""".format(
            self.stage_dir
        )

        self.assertThat(pc_file_content, Equals(expected_pc_file_content))

    def test_pc_files_correctly_prefixed_when_installed(self):
        pc_file = os.path.join("usr", "lib", "pkgconfig", "granite.pc")
        install_path = os.path.join(self.parts_dir, "stage-pc", "install")
        stage_pc_install = os.path.join(install_path, pc_file)
        stage_pc_stage = os.path.join(self.stage_dir, pc_file)

        # Run build
        project_config = self.make_snapcraft_project()
        lifecycle.execute(steps.BUILD, project_config)

        # Simulate a .pc file was installed
        os.makedirs(os.path.dirname(stage_pc_install))
        with open(stage_pc_install, "w") as f:
            f.write("prefix={}/usr\n".format(install_path))
            f.write("exec_prefix=${prefix}\n")
            f.write("libdir=${prefix}/lib\n")
            f.write("includedir=${prefix}/include\n")
            f.write("\n")
            f.write("Name: granite\n")
            f.write("Description: elementary's Application Framework\n")
            f.write("Version: 0.4\n")
            f.write("Libs: -L${libdir} -lgranite\n")
            f.write("Cflags: -I${includedir}/granite\n")
            f.write("Requires: cairo gee-0.8 glib-2.0 gio-unix-2.0 gobject-2.0\n")

        # Now we stage
        lifecycle.execute(steps.STAGE, project_config)

        with open(stage_pc_stage) as f:
            pc_file_content = f.read()
        expected_pc_file_content = """prefix={}/usr
exec_prefix=${{prefix}}
libdir=${{prefix}}/lib
includedir=${{prefix}}/include

Name: granite
Description: elementary's Application Framework
Version: 0.4
Libs: -L${{libdir}} -lgranite
Cflags: -I${{includedir}}/granite
Requires: cairo gee-0.8 glib-2.0 gio-unix-2.0 gobject-2.0
""".format(
            self.stage_dir
        )

        self.assertThat(pc_file_content, Equals(expected_pc_file_content))


class NextLastStepTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.handler = self.load_part("test_part")

    def test_pull(self):
        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        self.handler.pull()

        self.assertThat(self.handler.latest_step(), Equals(steps.PULL))
        self.assertThat(self.handler.next_step(), Equals(steps.BUILD))

    def test_build(self):
        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        self.handler.build()

        self.assertThat(self.handler.latest_step(), Equals(steps.BUILD))
        self.assertThat(self.handler.next_step(), Equals(steps.STAGE))

    def test_stage(self):
        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        self.handler.stage()

        self.assertThat(self.handler.latest_step(), Equals(steps.STAGE))
        self.assertThat(self.handler.next_step(), Equals(steps.PRIME))

    def test_prime(self):
        self.assertRaises(errors.NoLatestStepError, self.handler.latest_step)
        self.assertThat(self.handler.next_step(), Equals(steps.PULL))

        self.handler.prime()

        self.assertThat(self.handler.latest_step(), Equals(steps.PRIME))
        self.assertRaises(errors.NoNextStepError, self.handler.next_step)


class IsDirtyTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.handler = self.load_part("test-part")
        self.handler.makedirs()

    def test_prime_is_dirty(self):
        self.handler.plugin.options.snap = ["foo"]
        self.handler._part_properties = {"prime": ["foo"]}
        self.handler.mark_done(
            steps.PRIME,
            states.PrimeState(set(), set(), set(), self.handler._part_properties),
        )
        self.assertFalse(
            self.handler.is_clean(steps.PRIME), "Prime step was unexpectedly clean"
        )
        self.assertFalse(
            self.handler.is_dirty(steps.PRIME), "Strip step was unexpectedly dirty"
        )

        # Change the `snap` keyword-- thereby making the prime step dirty.
        self.handler.plugin.options.snap = ["bar"]
        self.handler._part_properties = {"prime": ["bar"]}
        self.assertFalse(
            self.handler.is_clean(steps.PRIME), "Strip step was unexpectedly clean"
        )
        self.assertTrue(
            self.handler.is_dirty(steps.PRIME), "Expected prime step to be dirty"
        )

    def test_prime_not_dirty_if_clean(self):
        self.assertTrue(
            self.handler.is_clean(steps.PRIME),
            "Expected vanilla handler to have clean prime step",
        )
        self.assertFalse(
            self.handler.is_dirty(steps.PRIME),
            "Expected vanilla handler to not have a dirty prime step",
        )

    def test_stage_is_dirty(self):
        self.handler = self.load_part("test-part", part_properties={"stage": ["foo"]})

        self.handler.mark_stage_done(set(), set())
        self.assertFalse(
            self.handler.is_clean(steps.STAGE), "Stage step was unexpectedly clean"
        )
        self.assertFalse(
            self.handler.is_dirty(steps.STAGE), "Stage step was unexpectedly dirty"
        )

        # Change the `stage` keyword-- thereby making the stage step dirty.
        self.handler = self.load_part("test-part", part_properties={"stage": ["bar"]})
        self.assertFalse(
            self.handler.is_clean(steps.STAGE), "Stage step was unexpectedly clean"
        )
        self.assertTrue(
            self.handler.is_dirty(steps.STAGE), "Expected stage step to be dirty"
        )

    def test_stage_not_dirty_if_clean(self):
        self.assertTrue(
            self.handler.is_clean(steps.STAGE),
            "Expected vanilla handler to have clean stage step",
        )
        self.assertFalse(
            self.handler.is_dirty(steps.STAGE),
            "Expected vanilla handler to not have a dirty stage step",
        )

    def test_build_is_dirty_from_options(self):
        self.useFixture(fixture_setup.FakePlugin("test-plugin", mocks.TestPlugin))
        self.handler = self.load_part(
            "test-part", "test-plugin", {"test-property": "foo"}
        )
        self.handler.mark_build_done()
        self.assertFalse(
            self.handler.is_clean(steps.BUILD), "Build step was unexpectedly clean"
        )
        self.assertFalse(
            self.handler.is_dirty(steps.BUILD), "Build step was unexpectedly dirty"
        )

        # Change `test-property`, thereby making the build step dirty.
        self.handler = self.load_part(
            "test-part", "test-plugin", {"test-property": "bar"}
        )
        self.assertFalse(
            self.handler.is_clean(steps.BUILD), "Build step was unexpectedly clean"
        )
        self.assertTrue(
            self.handler.is_dirty(steps.BUILD), "Expected build step to be dirty"
        )

    @patch.object(snapcraft.BasePlugin, "enable_cross_compilation")
    def test_build_is_dirty_from_project(self, mock_enable_cross_compilation):
        project = Project(target_deb_arch="amd64")
        self.handler = self.load_part("test-part", project=project)
        self.handler.mark_build_done()
        self.assertFalse(
            self.handler.is_clean(steps.BUILD), "Build step was unexpectedly clean"
        )
        self.assertFalse(
            self.handler.is_dirty(steps.BUILD), "Build step was unexpectedly dirty"
        )

        # Reload the plugin with new project options arch, thereby making it
        # dirty.
        project = Project(target_deb_arch="armhf")
        self.handler = self.load_part("test-part", project=project)
        self.assertFalse(
            self.handler.is_clean(steps.BUILD), "Build step was unexpectedly clean"
        )
        self.assertTrue(
            self.handler.is_dirty(steps.BUILD), "Expected build step to be dirty"
        )

    def test_build_not_dirty_if_clean(self):
        self.assertTrue(
            self.handler.is_clean(steps.BUILD),
            "Expected vanilla handler to have clean build step",
        )
        self.assertFalse(
            self.handler.is_dirty(steps.BUILD),
            "Expected vanilla handler to not have a dirty build step",
        )

    def test_pull_is_dirty_from_options(self):
        self.useFixture(fixture_setup.FakePlugin("test-plugin", mocks.TestPlugin))
        self.handler = self.load_part(
            "test-part", "test-plugin", {"test-property": "foo"}
        )
        self.handler.mark_pull_done()
        self.assertFalse(
            self.handler.is_clean(steps.PULL), "Pull step was unexpectedly clean"
        )
        self.assertFalse(
            self.handler.is_dirty(steps.PULL), "Pull step was unexpectedly dirty"
        )

        # Change `test-property`, thereby making the pull step dirty.
        self.handler = self.load_part(
            "test-part", "test-plugin", {"test-property": "bar"}
        )
        self.assertFalse(
            self.handler.is_clean(steps.PULL), "Pull step was unexpectedly clean"
        )
        self.assertTrue(
            self.handler.is_dirty(steps.PULL), "Expected pull step to be dirty"
        )

    @patch.object(snapcraft.BasePlugin, "enable_cross_compilation")
    def test_pull_is_dirty_from_project(self, mock_enable_cross_compilation):
        project = Project(target_deb_arch="amd64")
        self.handler = self.load_part("test-part", project=project)
        self.handler.mark_pull_done()
        self.assertFalse(
            self.handler.is_clean(steps.PULL), "Pull step was unexpectedly clean"
        )
        self.assertFalse(
            self.handler.is_dirty(steps.PULL), "Pull step was unexpectedly dirty"
        )

        # Reload the plugin with new project options arch, thereby making it
        # dirty.
        project = Project(target_deb_arch="armhf")
        self.handler = self.load_part("test-part", project=project)
        self.assertFalse(
            self.handler.is_clean(steps.PULL), "Pull step was unexpectedly clean"
        )
        self.assertTrue(
            self.handler.is_dirty(steps.PULL), "Expected pull step to be dirty"
        )

    def test_pull_not_dirty_if_clean(self):
        self.assertTrue(
            self.handler.is_clean(steps.PULL),
            "Expected vanilla handler to have clean pull step",
        )
        self.assertFalse(
            self.handler.is_dirty(steps.PULL),
            "Expected vanilla handler to not have a dirty pull step",
        )


class IsOutdatedTest(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.handler = self.load_part("test-part", part_properties=dict(source="."))
        self.handler.makedirs()

    def set_modified_time_later(self, target, reference):
        access_time = os.stat(reference).st_atime
        modified_time = os.stat(reference).st_atime
        os.utime(target, (access_time, modified_time + 1))

    def test_prime_is_outdated(self):
        self.handler.mark_prime_done(set(), set(), set(), set())
        self.assertFalse(
            self.handler.is_outdated(steps.PRIME),
            "Prime step was unexpectedly outdated",
        )

        # Now mark the stage state as done, but ensure it has a later
        # timestamp, thereby making the prime step out of date.
        prime_state_path = states.get_step_state_file(
            self.handler.plugin.statedir, steps.PRIME
        )
        stage_state_path = states.get_step_state_file(
            self.handler.plugin.statedir, steps.STAGE
        )
        open(stage_state_path, "w").close()
        self.set_modified_time_later(stage_state_path, prime_state_path)

        self.assertTrue(
            self.handler.is_outdated(steps.PRIME), "Expected prime step to be outdated"
        )

    def test_stage_is_outdated(self):
        self.handler.mark_stage_done(set(), set())

        self.assertFalse(
            self.handler.is_outdated(steps.STAGE),
            "Stage step was unexpectedly outdated",
        )

        # Now mark the build state as done, but ensure it has a later
        # timestamp, thereby making the stage step out of date.
        stage_state_path = states.get_step_state_file(
            self.handler.plugin.statedir, steps.STAGE
        )
        build_state_path = states.get_step_state_file(
            self.handler.plugin.statedir, steps.BUILD
        )
        open(build_state_path, "w").close()
        self.set_modified_time_later(build_state_path, stage_state_path)

        self.assertTrue(
            self.handler.is_outdated(steps.STAGE), "Expected stage step to be outdated"
        )

    def test_build_is_outdated(self):
        self.handler.mark_build_done()

        self.assertFalse(
            self.handler.is_outdated(steps.BUILD),
            "Build step was unexpectedly outdated",
        )

        # Now mark the pull state as done, but ensure it has a later
        # timestamp, thereby making the build step out of date.
        build_state_path = states.get_step_state_file(
            self.handler.plugin.statedir, steps.BUILD
        )
        pull_state_path = states.get_step_state_file(
            self.handler.plugin.statedir, steps.PULL
        )
        open(pull_state_path, "w").close()
        self.set_modified_time_later(pull_state_path, build_state_path)

        self.assertTrue(
            self.handler.is_outdated(steps.BUILD), "Expected build step to be outdated"
        )

    def test_pull_is_outdated(self):
        self.handler.mark_pull_done()

        self.assertFalse(
            self.handler.is_outdated(steps.PULL), "Pull step was unexpectedly outdated"
        )

        # Now update the source on disk, ensuring it has a later timestamp,
        # thereby making the pull step out of date.
        open("new_file", "w").close()
        pull_state_path = states.get_step_state_file(
            self.handler.plugin.statedir, steps.PULL
        )
        self.set_modified_time_later("new_file", pull_state_path)

        self.assertTrue(
            self.handler.is_outdated(steps.PULL), "Expected pull step to be outdated"
        )


class CollisionTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        tmpdirObject = tempfile.TemporaryDirectory()
        self.addCleanup(tmpdirObject.cleanup)
        tmpdir = tmpdirObject.name

        part1 = self.load_part("part1")
        part1.part_install_dir = tmpdir + "/install1"
        os.makedirs(part1.part_install_dir + "/a")
        open(part1.part_install_dir + "/a/1", mode="w").close()
        with open(part1.part_install_dir + "/file.pc", mode="w") as f:
            f.write("prefix={}\n".format(part1.part_install_dir))
            f.write("Name: File\n")

        part2 = self.load_part("part2")
        part2.part_install_dir = tmpdir + "/install2"
        os.makedirs(part2.part_install_dir + "/a")
        with open(part2.part_install_dir + "/1", mode="w") as f:
            f.write("1")
        open(part2.part_install_dir + "/2", mode="w").close()
        with open(part2.part_install_dir + "/a/2", mode="w") as f:
            f.write("a/2")
        with open(part2.part_install_dir + "/file.pc", mode="w") as f:
            f.write("prefix={}\n".format(part2.part_install_dir))
            f.write("Name: File\n")

        part3 = self.load_part("part3")
        part3.part_install_dir = tmpdir + "/install3"
        os.makedirs(part3.part_install_dir + "/a")
        os.makedirs(part3.part_install_dir + "/b")
        with open(part3.part_install_dir + "/1", mode="w") as f:
            f.write("2")
        with open(part2.part_install_dir + "/2", mode="w") as f:
            f.write("1")
        open(part3.part_install_dir + "/a/2", mode="w").close()

        part4 = self.load_part("part4")
        part4.part_install_dir = tmpdir + "/install4"
        os.makedirs(part4.part_install_dir)
        with open(part4.part_install_dir + "/file.pc", mode="w") as f:
            f.write("prefix={}\n".format(part4.part_install_dir))
            f.write("Name: ConflictFile\n")

        # Create a new part with a symlink that collides with part1's
        # non-symlink.
        part5 = self.load_part("part5")
        part5.part_install_dir = os.path.join(tmpdir, "install5")
        os.makedirs(part5.part_install_dir)
        os.symlink("foo", os.path.join(part5.part_install_dir, "a"))

        # Create a new part with a symlink that points to a different place
        # than part5's symlink.
        part6 = self.load_part("part6")
        part6.part_install_dir = os.path.join(tmpdir, "install6")
        os.makedirs(part6.part_install_dir)
        os.symlink("bar", os.path.join(part6.part_install_dir, "a"))

        self.part1 = part1
        self.part2 = part2
        self.part3 = part3
        self.part4 = part4
        self.part5 = part5
        self.part6 = part6

    def test_no_collisions(self):
        """No exception is expected as there are no collisions."""
        pluginhandler.check_for_collisions([self.part1, self.part2])

    def test_collisions_between_two_parts(self):
        raised = self.assertRaises(
            errors.SnapcraftPartConflictError,
            pluginhandler.check_for_collisions,
            [self.part1, self.part2, self.part3],
        )

        self.assertThat(raised.other_part_name, Equals("part2"))
        self.assertThat(raised.part_name, Equals("part3"))
        self.assertThat(raised.file_paths, Equals("    1\n    a/2"))

    def test_collisions_checks_symlinks(self):
        raised = self.assertRaises(
            errors.SnapcraftPartConflictError,
            pluginhandler.check_for_collisions,
            [self.part5, self.part6],
        )

        self.assertThat(
            str(raised),
            Contains(
                "Parts 'part5' and 'part6' have the following files, but with "
                "different contents:\n    a"
            ),
        )

    def test_collisions_not_both_symlinks(self):
        raised = self.assertRaises(
            errors.SnapcraftPartConflictError,
            pluginhandler.check_for_collisions,
            [self.part1, self.part5],
        )

        self.assertThat(
            str(raised),
            Contains(
                "Parts 'part1' and 'part5' have the following files, but with "
                "different contents:\n    a"
            ),
        )

    def test_collisions_between_two_parts_pc_files(self):
        raised = self.assertRaises(
            errors.SnapcraftPartConflictError,
            pluginhandler.check_for_collisions,
            [self.part1, self.part4],
        )

        self.assertThat(raised.other_part_name, Equals("part1"))
        self.assertThat(raised.part_name, Equals("part4"))
        self.assertThat(raised.file_paths, Equals("    file.pc"))

    def test_collision_with_part_not_built(self):
        part_built = self.load_part(
            "part_built", part_properties={"stage": ["collision"]}
        )
        part_built.part_install_dir = "part_built_install"
        # a part built has the stage file in the installdir.
        os.makedirs(part_built.part_install_dir)
        open(os.path.join(part_built.part_install_dir, "collision"), "w").close()
        part_not_built = self.load_part(
            "part_not_built", part_properties={"stage": ["collision"]}
        )
        part_not_built.part_install_dir = "part_not_built_install"
        # a part not built doesn't have the stage file in the installdir.
        pluginhandler.check_for_collisions([part_built, part_not_built])


class StagePackagesTestCase(unit.TestCase):
    def test_missing_stage_package_raises_exception(self):
        fake_repo = Mock()
        fake_repo.fetch_stage_packages.side_effect = repo.errors.PackageNotFoundError(
            "non-existing"
        )
        part = self.load_part(
            "stage-test",
            part_properties={"stage-packages": ["non-existing"]},
            stage_packages_repo=fake_repo,
        )

        raised = self.assertRaises(errors.StagePackageDownloadError, part.prepare_pull)

        self.assertThat(raised.part_name, Equals("stage-test"))
        self.assertThat(
            raised.message, Equals("The package 'non-existing' was not found.")
        )


class FilesetsTestCase(unit.TestCase):
    def test_combine_filesets_explicit_wildcard(self):
        fileset_1 = ["a", "b"]
        fileset_2 = ["*"]

        expected_fileset = ["a", "b"]
        combined_fileset = pluginhandler._combine_filesets(fileset_1, fileset_2)
        self.assertThat(set(combined_fileset), Equals(set(expected_fileset)))

    def test_combine_filesets_implicit_wildcard(self):
        fileset_1 = ["a", "b"]
        fileset_2 = ["-c"]

        expected_fileset = ["a", "-c", "b"]
        combined_fileset = pluginhandler._combine_filesets(fileset_1, fileset_2)
        self.assertThat(set(combined_fileset), Equals(set(expected_fileset)))

    def test_combine_filesets_no_wildcard(self):
        fileset_1 = ["a", "b"]
        fileset_2 = ["a"]

        expected_fileset = ["a"]
        combined_fileset = pluginhandler._combine_filesets(fileset_1, fileset_2)
        self.assertThat(set(combined_fileset), Equals(set(expected_fileset)))

    def test_combine_filesets_with_contradiciton(self):
        fileset_1 = ["-a"]
        fileset_2 = ["a"]

        raised = self.assertRaises(
            errors.PrimeFileConflictError,
            pluginhandler._combine_filesets,
            fileset_1,
            fileset_2,
        )
        self.assertThat(raised.fileset, Equals({"a"}))

    def test_get_includes(self):
        fileset = ["-a", "b"]
        expected_includes = ["b"]

        includes = pluginhandler._get_includes(fileset)
        self.assertThat(set(includes), Equals(set(expected_includes)))

    def test_get_excludes(self):
        fileset = ["-a", "b"]
        expected_excludes = ["a"]

        excludes = pluginhandler._get_excludes(fileset)
        self.assertThat(set(excludes), Equals(set(expected_excludes)))


class SourcesTestCase(unit.TestCase):
    def test_do_not_follow_links(self):
        properties = dict(source=".")
        handler = self.load_part("test-part", part_properties=properties)

        # Create a file and a symlink to it
        open("file", mode="w").close()
        os.symlink("file", "symlinkfile")

        # Create a directory and a symlink to it
        os.mkdir("dir")
        os.symlink("dir", "symlinkdir")

        handler.pull()
        handler.build()

        # Make sure this is still a link
        build_file_path = os.path.join(handler.part_build_dir, "file")
        build_symlinkfile_path = os.path.join(handler.part_build_dir, "symlinkfile")

        self.assertTrue(os.path.isfile(build_file_path))
        self.assertTrue(os.path.islink(build_symlinkfile_path))

        build_dir_path = os.path.join(handler.part_build_dir, "dir")
        build_symlinkdir_path = os.path.join(handler.part_build_dir, "symlinkdir")

        self.assertTrue(os.path.isdir(build_dir_path))
        self.assertTrue(os.path.isdir(build_symlinkdir_path))

    def test_pull_ignores_snapcraft_files_in_source_dir(self):
        properties = dict(source=".")
        handler = self.load_part("test-part", part_properties=properties)

        open("my-snap.snap", "w").close()
        open("my-snap", "w").close()

        handler.pull()

        for file_ in common.SNAPCRAFT_FILES:
            self.assertFalse(
                os.path.exists(os.path.join(handler.part_source_dir, file_))
            )
        self.assertThat(
            os.path.join(handler.part_source_dir, "my-snap.snap"), Not(FileExists())
        )

        # Make sure we don't filter things out incorrectly
        self.assertThat(os.path.join(handler.part_source_dir, "my-snap"), FileExists())

    def test_source_with_unrecognized_source_must_raise_exception(self):
        properties = dict(source="unrecognized://test_source")

        self.assertRaises(
            SnapcraftSourceUnhandledError,
            self.load_part,
            "test-part",
            part_properties=properties,
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from unittest import mock

import fixtures
import pytest
from testtools.matchers import Equals, HasLength

import snapcraft
from snapcraft.internal import errors, meta
from snapcraft.plugins.v1 import kbuild

from . import PluginsV1BaseTestCase


class KBuildPluginTestCase(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        class Options:
            build_parameters = []
            kconfigfile = None
            kconfigflavour = None
            kdefconfig = []
            kconfigs = []
            build_attributes = []

        self.options = Options()

    def test_schema(self):
        schema = kbuild.KBuildPlugin.schema()

        properties = schema["properties"]
        self.assertThat(properties["kdefconfig"]["type"], Equals("array"))
        self.assertThat(properties["kdefconfig"]["default"], Equals(["defconfig"]))

        self.assertThat(properties["kconfigfile"]["type"], Equals("string"))
        self.assertThat(properties["kconfigfile"]["default"], Equals(None))

        self.assertThat(properties["kconfigflavour"]["type"], Equals("string"))
        self.assertThat(properties["kconfigflavour"]["default"], Equals(None))

        self.assertThat(properties["kconfigs"]["type"], Equals("array"))
        self.assertThat(properties["kconfigs"]["default"], Equals([]))
        self.assertThat(properties["kconfigs"]["minitems"], Equals(1))
        self.assertThat(properties["kconfigs"]["items"]["type"], Equals("string"))
        self.assertTrue(properties["kconfigs"]["uniqueItems"])

    def test_get_build_properties(self):
        expected_build_properties = [
            "kdefconfig",
            "kconfigfile",
            "kconfigflavour",
            "kconfigs",
        ]
        resulting_build_properties = kbuild.KBuildPlugin.get_build_properties()

        self.assertThat(
            resulting_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, resulting_build_properties)

    @mock.patch("subprocess.check_call")
    @mock.patch.object(kbuild.KBuildPlugin, "run")
    def test_build_with_kconfigfile(self, run_mock, check_call_mock):
        self.options.kconfigfile = "config"
        with open(self.options.kconfigfile, "w") as f:
            f.write("ACCEPT=y\n")

        plugin = kbuild.KBuildPlugin("test-part", self.options, self.project)

        os.makedirs(plugin.builddir)

        plugin.build()

        self.assertThat(check_call_mock.call_count, Equals(1))
        check_call_mock.assert_has_calls(
            [mock.call('yes "" | make -j2 oldconfig', shell=True, cwd=plugin.builddir)]
        )

        self.assertThat(run_mock.call_count, Equals(2))
        run_mock.assert_has_calls(
            [
                mock.call(["make", "-j2"]),
                mock.call(
                    [
                        "make",
                        "-j2",
                        "CONFIG_PREFIX={}".format(plugin.installdir),
                        "install",
                    ]
                ),
            ]
        )

        config_file = os.path.join(plugin.builddir, ".config")
        self.assertTrue(os.path.exists(config_file))

        with open(config_file) as f:
            config_contents = f.read()

        self.assertThat(config_contents, Equals("ACCEPT=y\n"))

    @mock.patch("subprocess.check_call")
    @mock.patch.object(kbuild.KBuildPlugin, "run")
    def test_build_verbose_with_kconfigfile(self, run_mock, check_call_mock):
        fake_logger = fixtures.FakeLogger(level=logging.DEBUG)
        self.useFixture(fake_logger)

        self.options.kconfigfile = "config"
        with open(self.options.kconfigfile, "w") as f:
            f.write("ACCEPT=y\n")

        plugin = kbuild.KBuildPlugin("test-part", self.options, self.project)

        os.makedirs(plugin.builddir)

        plugin.build()

        self.assertThat(check_call_mock.call_count, Equals(1))
        check_call_mock.assert_has_calls(
            [
                mock.call(
                    'yes "" | make -j2 V=1 oldconfig', shell=True, cwd=plugin.builddir
                )
            ]
        )

        self.assertThat(run_mock.call_count, Equals(2))
        run_mock.assert_has_calls(
            [
                mock.call(["make", "-j2", "V=1"]),
                mock.call(
                    [
                        "make",
                        "-j2",
                        "V=1",
                        "CONFIG_PREFIX={}".format(plugin.installdir),
                        "install",
                    ]
                ),
            ]
        )

        config_file = os.path.join(plugin.builddir, ".config")
        self.assertTrue(os.path.exists(config_file))

        with open(config_file) as f:
            config_contents = f.read()

        self.assertThat(config_contents, Equals("ACCEPT=y\n"))

    @mock.patch("subprocess.check_call")
    @mock.patch.object(kbuild.KBuildPlugin, "run")
    def test_build_with_kconfigfile_and_kconfigs(self, run_mock, check_call_mock):
        self.options.kconfigfile = "config"
        self.options.kconfigs = ["SOMETHING=y", "ACCEPT=n"]

        with open(self.options.kconfigfile, "w") as f:
            f.write("ACCEPT=y\n")

        plugin = kbuild.KBuildPlugin("test-part", self.options, self.project)

        os.makedirs(plugin.builddir)

        plugin.build()

        self.assertThat(check_call_mock.call_count, Equals(1))
        check_call_mock.assert_has_calls(
            [mock.call('yes "" | make -j2 oldconfig', shell=True, cwd=plugin.builddir)]
        )

        self.assertThat(run_mock.call_count, Equals(2))
        run_mock.assert_has_calls(
            [
                mock.call(["make", "-j2"]),
                mock.call(
                    [
                        "make",
                        "-j2",
                        "CONFIG_PREFIX={}".format(plugin.installdir),
                        "install",
                    ]
                ),
            ]
        )

        config_file = os.path.join(plugin.builddir, ".config")
        self.assertTrue(os.path.exists(config_file))

        with open(config_file) as f:
            config_contents = f.read()

        expected_config = """SOMETHING=y
ACCEPT=n

ACCEPT=y

SOMETHING=y
ACCEPT=n
"""
        self.assertThat(config_contents, Equals(expected_config))

        # Finally, ensure that the original kconfigfile was not modified (it's back in
        # the source tree)
        with open(self.options.kconfigfile, "r") as f:
            self.assertThat(f.read(), Equals("ACCEPT=y\n"))

    @mock.patch("subprocess.check_call")
    @mock.patch.object(kbuild.KBuildPlugin, "run")
    def test_build_with_defconfig_and_kconfigs(self, run_mock, check_call_mock):
        self.options.kdefconfig = ["defconfig"]
        self.options.kconfigs = ["SOMETHING=y", "ACCEPT=n"]

        plugin = kbuild.KBuildPlugin("test-part", self.options, self.project)

        config_file = os.path.join(plugin.builddir, ".config")

        def fake_defconfig(*args, **kwargs):
            if os.path.exists(config_file):
                return
            with open(config_file, "w") as f:
                f.write("ACCEPT=y\n")

        run_mock.side_effect = fake_defconfig

        os.makedirs(plugin.builddir)

        plugin.build()

        self.assertThat(check_call_mock.call_count, Equals(1))
        check_call_mock.assert_has_calls(
            [mock.call('yes "" | make -j2 oldconfig', shell=True, cwd=plugin.builddir)]
        )

        self.assertThat(run_mock.call_count, Equals(3))
        run_mock.assert_has_calls(
            [
                mock.call(["make", "-j2"]),
                mock.call(
                    [
                        "make",
                        "-j2",
                        "CONFIG_PREFIX={}".format(plugin.installdir),
                        "install",
                    ]
                ),
            ]
        )

        self.assertTrue(os.path.exists(config_file))

        with open(config_file) as f:
            config_contents = f.read()

        expected_config = """SOMETHING=y
ACCEPT=n

ACCEPT=y

SOMETHING=y
ACCEPT=n
"""
        self.assertThat(config_contents, Equals(expected_config))

    def test_unsupported_base(self):
        self.project._snap_meta.base = "unsupported-base"

        raised = self.assertRaises(
            errors.PluginBaseError,
            kbuild.KBuildPlugin,
            "test-part",
            self.options,
            self.project,
        )

        self.assertThat(raised.part_name, Equals("test-part"))
        self.assertThat(raised.base, Equals("unsupported-base"))


@pytest.mark.parametrize("deb_arch", ["armhf", "arm64", "i386", "ppc64el"])
@mock.patch("subprocess.check_call")
def test_cross_compile(mock_check_call, monkeypatch, mock_run, deb_arch):
    monkeypatch.setattr(snapcraft.project.Project, "is_cross_compiling", True)

    class Options:
        build_parameters = []
        kconfigfile = None
        kconfigflavour = None
        kdefconfig = []
        kconfigs = []
        build_attributes = []

    project = snapcraft.project.Project(target_deb_arch=deb_arch)
    project._snap_meta = meta.snap.Snap(name="test-snap", base="core18")

    plugin = kbuild.KBuildPlugin("test-part", Options(), project)
    plugin.enable_cross_compilation()

    plugin.build()
    mock_run.assert_has_calls(
        [
            mock.call(
                [
                    "make",
                    "-j1",
                    f"ARCH={project.kernel_arch}",
                    f"CROSS_COMPILE={project.cross_compiler_prefix}",
                    mock.ANY,
                ]
            )
        ]
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import requests
import urllib3
from unittest import mock

from snapcraft.storeapi.http_clients import errors


def _fake_error_response(status_code, reason):
    response = mock.Mock()
    response.status_code = status_code
    response.reason = reason
    return response


class TestSnapcraftException:
    scenarios = (
        (
            "InvalidCredentialsError",
            {
                "exception_class": errors.InvalidCredentialsError,
                "kwargs": {"message": "macaroon expired"},
                "expected_message": (
                    "Invalid credentials: macaroon expired. "
                    'Have you run "snapcraft login"?'
                ),
            },
        ),
        (
            "StoreAuthenticationError",
            {
                "exception_class": errors.StoreAuthenticationError,
                "kwargs": {"message": "invalid password"},
                "expected_message": ("Authentication error: invalid password"),
            },
        ),
        (
            "StoreNetworkError generic error",
            {
                "exception_class": errors.StoreNetworkError,
                "kwargs": {
                    "exception": requests.exceptions.ConnectionError("bad error")
                },
                "expected_message": "There seems to be a network error: bad error",
            },
        ),
        (
            "StoreNetworkError max retry error",
            {
                "exception_class": errors.StoreNetworkError,
                "kwargs": {
                    "exception": requests.exceptions.ConnectionError(
                        urllib3.exceptions.MaxRetryError(
                            pool="test-pool", url="test-url"
                        )
                    )
                },
                "expected_message": (
                    "There seems to be a network error: maximum retries exceeded "
                    "trying to reach the store.\n"
                    "Check your network connection, and check the store status at "
                    "https://status.snapcraft.io/"
                ),
            },
        ),
        (
            "StoreServerError 500",
            {
                "exception_class": errors.StoreServerError,
                "kwargs": {
                    "response": _fake_error_response(500, "internal server error")
                },
                "expected_message": (
                    "The Snap Store encountered an error while processing your "
                    "request: internal server error (code 500).\nThe operational "
                    "status of the Snap Store can be checked at "
                    "https://status.snapcraft.io/"
                ),
            },
        ),
        (
            "StoreServerError 501",
            {
                "exception_class": errors.StoreServerError,
                "kwargs": {"response": _fake_error_response(501, "not implemented")},
                "expected_message": (
                    "The Snap Store encountered an error while processing your "
                    "request: not implemented (code 501).\nThe operational "
                    "status of the Snap Store can be checked at "
                    "https://status.snapcraft.io/"
                ),
            },
        ),
    )

    def test_error_formatting(self, exception_class, expected_message, kwargs):
        assert str(exception_class(**kwargs)) == expected_message



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import copy
import io
import os
import platform
import shutil
import subprocess
import sys
import tempfile
import threading
import urllib.parse
from functools import partial
from unittest import mock

import fixtures
import xdg

from tests import fake_servers
from tests.fake_servers import api, search, upload
from tests.subprocess_utils import call, call_with_output

# we do not want snapcraft imports for the integration tests
try:
    from snapcraft import yaml_utils
except ImportError:
    import yaml as yaml_utils  # type: ignore


class TempCWD(fixtures.Fixture):
    def __init__(self, rootdir=None):
        super().__init__()
        if rootdir is None and "TMPDIR" in os.environ:
            rootdir = os.environ.get("TMPDIR")
        self.rootdir = rootdir
        self._data_path = os.getenv("SNAPCRAFT_TEST_KEEP_DATA_PATH", None)

    def setUp(self):
        """Create a temporary directory an cd into it for the test duration."""
        super().setUp()
        if self._data_path:
            os.makedirs(self._data_path)
            self.path = self._data_path
        else:
            self.path = tempfile.mkdtemp(dir=self.rootdir)
        current_dir = os.getcwd()
        self.addCleanup(os.chdir, current_dir)
        if not self._data_path:
            self.addCleanup(shutil.rmtree, self.path, ignore_errors=True)
        os.chdir(self.path)


class TempXDG(fixtures.Fixture):
    """Isolate a test from xdg so a private temp config is used."""

    def __init__(self, path):
        super().setUp()
        self.path = path

    def setUp(self):
        super().setUp()
        patcher = mock.patch(
            "xdg.BaseDirectory.xdg_config_home", new=os.path.join(self.path, ".config")
        )
        patcher.start()
        self.addCleanup(patcher.stop)
        patcher = mock.patch(
            "xdg.BaseDirectory.xdg_data_home", new=os.path.join(self.path, ".local")
        )
        patcher.start()
        self.addCleanup(patcher.stop)
        patcher = mock.patch(
            "xdg.BaseDirectory.xdg_cache_home", new=os.path.join(self.path, ".cache")
        )
        patcher.start()
        self.addCleanup(patcher.stop)

        patcher_dirs = mock.patch(
            "xdg.BaseDirectory.xdg_config_dirs", new=[xdg.BaseDirectory.xdg_config_home]
        )
        patcher_dirs.start()
        self.addCleanup(patcher_dirs.stop)

        patcher_dirs = mock.patch(
            "xdg.BaseDirectory.xdg_data_dirs", new=[xdg.BaseDirectory.xdg_data_home]
        )
        patcher_dirs.start()
        self.addCleanup(patcher_dirs.stop)

        self.useFixture(
            fixtures.EnvironmentVariable(
                "XDG_CONFIG_HOME", os.path.join(self.path, ".config")
            )
        )
        self.useFixture(
            fixtures.EnvironmentVariable(
                "XDG_DATA_HOME", os.path.join(self.path, ".local")
            )
        )
        self.useFixture(
            fixtures.EnvironmentVariable(
                "XDG_CACHE_HOME", os.path.join(self.path, ".cache")
            )
        )


class CleanEnvironment(fixtures.Fixture):
    def setUp(self):
        super().setUp()

        current_environment = copy.deepcopy(os.environ)
        os.environ.clear()

        self.addCleanup(os.environ.update, current_environment)


class _FakeStdout(io.StringIO):
    """A fake stdout using StringIO implementing the missing fileno attrib."""

    def fileno(self):
        return 1


class _FakeStderr(io.StringIO):
    """A fake stderr using StringIO implementing the missing fileno attrib."""

    def fileno(self):
        return 2


class _FakeTerminalSize:
    def __init__(self, columns=80):
        self.columns = columns


class FakeTerminal(fixtures.Fixture):
    def __init__(self, columns=80, isatty=True):
        self.columns = columns
        self.isatty = isatty

    def _setUp(self):
        patcher = mock.patch("shutil.get_terminal_size")
        mock_terminal_size = patcher.start()
        mock_terminal_size.return_value = _FakeTerminalSize(self.columns)
        self.addCleanup(patcher.stop)

        patcher = mock.patch("sys.stdout", new_callable=_FakeStdout)
        self.mock_stdout = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("sys.stderr", new_callable=_FakeStderr)
        self.mock_stderr = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("os.isatty")
        mock_isatty = patcher.start()
        mock_isatty.return_value = self.isatty
        self.addCleanup(patcher.stop)

    def getvalue(self, stderr=False):
        if stderr:
            return self.mock_stderr.getvalue()
        else:
            return self.mock_stdout.getvalue()


class FakePartsWiki(fixtures.Fixture):
    def setUp(self):
        super().setUp()

        self.fake_parts_wiki_fixture = FakePartsWikiRunning()
        self.useFixture(self.fake_parts_wiki_fixture)
        self.useFixture(fixtures.EnvironmentVariable("no_proxy", "localhost,127.0.0.1"))


class FakePartsWikiWithSlashes(fixtures.Fixture):
    def setUp(self):
        super().setUp()

        self.fake_parts_wiki_with_slashes_fixture = FakePartsWikiWithSlashesRunning()
        self.useFixture(self.fake_parts_wiki_with_slashes_fixture)
        self.useFixture(fixtures.EnvironmentVariable("no_proxy", "localhost,127.0.0.1"))


class FakePartsWikiOrigin(fixtures.Fixture):
    def setUp(self):
        super().setUp()

        self.fake_parts_wiki_origin_fixture = FakePartsWikiOriginRunning()
        self.useFixture(self.fake_parts_wiki_origin_fixture)
        self.useFixture(fixtures.EnvironmentVariable("no_proxy", "localhost,127.0.0.1"))


class FakeParts(fixtures.Fixture):
    def setUp(self):
        super().setUp()

        self.fake_parts_server_fixture = FakePartsServerRunning()
        self.useFixture(self.fake_parts_server_fixture)
        self.useFixture(
            fixtures.EnvironmentVariable(
                "SNAPCRAFT_PARTS_URI",
                urllib.parse.urljoin(self.fake_parts_server_fixture.url, "parts.yaml"),
            )
        )
        self.useFixture(fixtures.EnvironmentVariable("no_proxy", "localhost,127.0.0.1"))


class FakeStore(fixtures.Fixture):
    def setUp(self):
        super().setUp()
        # In case the variable was not set or it was empty.
        self.useFixture(fixtures.EnvironmentVariable("TEST_STORE", "fake"))

        self.needs_refresh = False

        self.fake_sso_server_fixture = FakeSSOServerRunning(self)
        self.useFixture(self.fake_sso_server_fixture)
        self.useFixture(
            fixtures.EnvironmentVariable(
                "UBUNTU_ONE_SSO_URL", self.fake_sso_server_fixture.url
            )
        )

        self.useFixture(fixtures.EnvironmentVariable("STORE_RETRIES", "1"))
        self.useFixture(fixtures.EnvironmentVariable("STORE_BACKOFF", "0"))

        self.fake_store_upload_server_fixture = FakeStoreUploadServerRunning()
        self.useFixture(self.fake_store_upload_server_fixture)
        self.useFixture(
            fixtures.EnvironmentVariable(
                "STORE_UPLOAD_URL", self.fake_store_upload_server_fixture.url,
            )
        )

        self.fake_store_api_server_fixture = FakeStoreAPIServerRunning(self)
        self.useFixture(self.fake_store_api_server_fixture)
        self.useFixture(
            fixtures.EnvironmentVariable(
                "STORE_DASHBOARD_URL", self.fake_store_api_server_fixture.url
            )
        )

        self.fake_store_search_server_fixture = FakeStoreSearchServerRunning()
        self.useFixture(self.fake_store_search_server_fixture)
        self.useFixture(
            fixtures.EnvironmentVariable(
                "STORE_API_URL", self.fake_store_search_server_fixture.url,
            )
        )

        self.useFixture(fixtures.EnvironmentVariable("no_proxy", "localhost,127.0.0.1"))


class FakeServerRunning(fixtures.Fixture):
    # fake_server needs to be set by implementing classes

    def setUp(self):
        super().setUp()
        self._start_fake_server()

    def _start_fake_server(self):
        server_address = ("", 0)
        self.server = self.fake_server(server_address)
        server_thread = threading.Thread(target=self.server.serve_forever)
        server_thread.start()
        self.addCleanup(self._stop_fake_server, server_thread)
        self.url = "http://localhost:{}/".format(self.server.server_port)

    def _stop_fake_server(self, thread):
        self.server.shutdown()
        self.server.socket.close()
        thread.join()


class FakePartsWikiOriginRunning(FakeServerRunning):

    fake_server = fake_servers.FakePartsWikiOriginServer


class FakePartsWikiRunning(FakeServerRunning):

    fake_server = fake_servers.FakePartsWikiServer


class FakePartsWikiWithSlashesRunning(FakeServerRunning):

    fake_server = fake_servers.FakePartsWikiWithSlashesServer


class FakePartsServerRunning(FakeServerRunning):

    fake_server = fake_servers.FakePartsServer


class FakeSSOServerRunning(FakeServerRunning):
    def __init__(self, fake_store):
        super().__init__()
        self.fake_server = partial(fake_servers.FakeSSOServer, fake_store)


class FakeStoreUploadServerRunning(FakeServerRunning):

    fake_server = upload.FakeStoreUploadServer


class FakeStoreAPIServerRunning(FakeServerRunning):
    def __init__(self, fake_store):
        super().__init__()
        self.fake_server = partial(api.FakeStoreAPIServer, fake_store)


class FakeStoreSearchServerRunning(FakeServerRunning):

    fake_server = search.FakeStoreSearchServer


class StagingStore(fixtures.Fixture):
    def setUp(self):
        super().setUp()
        self.useFixture(
            fixtures.EnvironmentVariable(
                "STORE_DASHBOARD_URL", "https://dashboard.staging.snapcraft.io/",
            )
        )
        self.useFixture(
            fixtures.EnvironmentVariable(
                "STORE_UPLOAD_URL", "https://upload.apps.staging.ubuntu.com/",
            )
        )
        self.useFixture(
            fixtures.EnvironmentVariable(
                "UBUNTU_ONE_SSO_URL", "https://login.staging.ubuntu.com/"
            )
        )
        self.useFixture(
            fixtures.EnvironmentVariable(
                "STORE_API_URL", "https://api.staging.snapcraft.io/"
            )
        )


class TestStore(fixtures.Fixture):
    def setUp(self):
        super().setUp()
        test_store = os.getenv("TEST_STORE") or "fake"
        if test_store == "fake":
            self.useFixture(FakeStore())
            self.register_count_limit = 10
            self.reserved_snap_name = "test-reserved-snap-name"
            self.already_owned_snap_name = "test-already-owned-snap-name"
        elif test_store == "staging":
            self.useFixture(StagingStore())
            self.register_count_limit = 100
            self.reserved_snap_name = "snapcraft-reserved-snap-name"
        elif test_store == "production":
            # Use the default server URLs
            self.register_count_limit = 10
            self.reserved_snap_name = "bash"
        else:
            raise ValueError("Unknown test store option: {}".format(test_store))

        # Do not change this email address. If you use a different address, it
        # will pollute KPIs, so always notify the store team if you need to use
        # a different value.
        self.user_email = (
            os.getenv("TEST_USER_EMAIL") or "snapcraft-test+user@canonical.com"
        )
        self.test_track_snap = os.getenv(
            "TEST_SNAP_WITH_TRACKS", "test-snapcraft-tracks"
        )
        if test_store == "fake":
            self.user_password = "test correct password"
        else:
            self.user_password = os.getenv("TEST_USER_PASSWORD")


class GitRepo(fixtures.Fixture):
    """Create a git repo in the current directory"""

    def setUp(self):
        super().setUp()
        name = "git-source"  # must match what the tests expect

        def _add_and_commit_file(path, filename, contents=None, message=None):
            if not contents:
                contents = filename
            if not message:
                message = filename

            with open(os.path.join(path, filename), "w") as fp:
                fp.write(contents)

            call(["git", "-C", name, "add", filename])
            call(["git", "-C", name, "commit", "-am", message])

        os.makedirs(name)
        call(["git", "-C", name, "init"])
        call(["git", "-C", name, "config", "user.name", "Test User"])
        call(["git", "-C", name, "config", "user.email", "testuser@example.com"])

        _add_and_commit_file(name, "testing")
        call(["git", "-C", name, "branch", "test-branch"])

        _add_and_commit_file(name, "testing-2")
        call(["git", "-C", name, "tag", "feature-tag"])

        _add_and_commit_file(name, "testing-3")

        self.commit = call_with_output(["git", "-C", name, "rev-parse", "HEAD"])


@contextlib.contextmanager
def return_to_cwd():
    cwd = os.getcwd()
    try:
        yield
    finally:
        os.chdir(cwd)


class BzrRepo(fixtures.Fixture):
    def __init__(self, name):
        self.name = name

    def setUp(self):
        super().setUp()

        bzr_home = self.useFixture(fixtures.TempDir()).path
        self.useFixture(fixtures.EnvironmentVariable("BZR_HOME", bzr_home))
        self.useFixture(
            fixtures.EnvironmentVariable(
                "BZR_EMAIL", "Test User <test.user@example.com>"
            )
        )

        with return_to_cwd():
            os.makedirs(self.name)
            os.chdir(self.name)
            call(["bzr", "init"])
            with open("testing", "w") as fp:
                fp.write("testing")

            call(["bzr", "add", "testing"])
            call(["bzr", "commit", "-m", "testing"])
            call(["bzr", "tag", "feature-tag"])
            revno = call_with_output(["bzr", "revno"])

            self.commit = revno


class SvnRepo(fixtures.Fixture):
    def __init__(self, name):
        self.name = name

    def setUp(self):
        super().setUp()

        working_tree = "svn-repo"
        call(["svnadmin", "create", self.name])
        call(
            [
                "svn",
                "checkout",
                "file://{}".format(os.path.join(os.getcwd(), self.name)),
                working_tree,
            ]
        )

        with return_to_cwd():
            os.chdir(working_tree)
            with open("testing", "w") as fp:
                fp.write("testing")

            call(["svn", "add", "testing"])
            call(["svn", "commit", "-m", "svn testing"])
            revno = "1"

            self.commit = revno


class HgRepo(fixtures.Fixture):
    def __init__(self, name):
        self.name = name

    def setUp(self):
        super().setUp()

        with return_to_cwd():
            os.makedirs(self.name)
            os.chdir(self.name)
            call(["hg", "init"])
            with open("testing", "w") as fp:
                fp.write("testing")

            call(["hg", "add", "testing"])
            call(
                [
                    "hg",
                    "commit",
                    "-m",
                    "testing",
                    "-u",
                    "Test User <test.user@example.com>",
                ]
            )
            call(
                ["hg", "tag", "feature-tag", "-u", "Test User <test.user@example.com>"]
            )
            revno = call_with_output(["hg", "id"]).split()[0]

            self.commit = revno


class WithoutSnapInstalled(fixtures.Fixture):
    """Assert that a snap is not installed and remove it on clean up.

    :raises: AssertionError: if the snap is installed when this fixture is
        set up.
    """

    def __init__(self, snap_name: str) -> None:
        super().__init__()
        self.snap_name = snap_name.split("/")[0]

    def setUp(self) -> None:
        super().setUp()
        with contextlib.suppress(subprocess.CalledProcessError):
            subprocess.check_call(["snap", "list", self.snap_name])
            raise AssertionError(
                "This test cannot run if you already have the {snap!r} snap "
                "installed. Please uninstall it by running "
                "'sudo snap remove {snap}'.".format(snap=self.snap_name)
            )

        self.addCleanup(self._remove_snap)

    def _remove_snap(self) -> None:
        try:
            subprocess.check_output(
                ["sudo", "snap", "remove", self.snap_name], stderr=subprocess.STDOUT
            )
        except subprocess.CalledProcessError as e:
            RuntimeError("unable to remove {!r}: {}".format(self.snap_name, e.output))


class SnapcraftYaml(fixtures.Fixture):
    def __init__(  # noqa: C901
        self,
        path,
        name="test-snap",
        base="core18",
        build_base=None,
        version="test-version",
        summary="test-summary",
        description="test-description",
        confinement="strict",
        architectures=None,
        apps=None,
        hooks=None,
        parts=None,
        type="app",
    ):
        super().__init__()

        if apps is None:
            apps = dict()

        if parts is None:
            parts = dict()

        self.path = path
        self.data = {
            "confinement": confinement,
            "parts": parts,
            "apps": apps,
            "type": type,
        }
        if name is not None:
            self.data["name"] = name
        if version is not None:
            self.data["version"] = version
        if summary is not None:
            self.data["summary"] = summary
        if description is not None:
            self.data["description"] = description
        if architectures is not None:
            self.data["architectures"] = architectures
        if base is not None:
            self.data["base"] = base
        if build_base is not None:
            self.data["build-base"] = build_base
        if hooks is not None:
            self.data["hooks"] = hooks

    def update_part(self, name, data):
        part = {name: data}
        self.data["parts"].update(part)
        self.write_snapcraft_yaml()

    def update_app(self, name, data):
        app = {name: data}
        self.data["apps"].update(app)
        self.write_snapcraft_yaml()

    def write_snapcraft_yaml(self):
        self.snapcraft_yaml_file_path = os.path.join(
            self.path, "snap", "snapcraft.yaml"
        )
        os.makedirs(os.path.join(self.path, "snap"), exist_ok=True)
        with open(self.snapcraft_yaml_file_path, "w") as snapcraft_yaml_file:
            yaml_utils.dump(self.data, stream=snapcraft_yaml_file)

    def _setUp(self):
        super()._setUp()
        self.write_snapcraft_yaml()


class SharedCache(fixtures.Fixture):
    def __init__(self, name) -> None:
        super().__init__()
        self.name = name

    def setUp(self) -> None:
        super().setUp()
        shared_cache_dir = os.path.join(
            tempfile.gettempdir(), "snapcraft_test_cache_{}".format(self.name)
        )
        os.makedirs(shared_cache_dir, exist_ok=True)
        self.useFixture(
            fixtures.EnvironmentVariable("XDG_CACHE_HOME", shared_cache_dir)
        )
        patcher = mock.patch(
            "xdg.BaseDirectory.xdg_cache_home", new=os.path.join(shared_cache_dir)
        )
        patcher.start()
        self.addCleanup(patcher.stop)


class FakeBaseEnvironment(fixtures.Fixture):

    _LINKER_FOR_ARCH = dict(
        armv7l="lib/ld-linux-armhf.so.3",
        aarch64="lib/ld-linux-aarch64.so.1",
        i686="lib/ld-linux.so.2",
        ppc64le="lib64/ld64.so.2",
        ppc="lib/ld-linux.so.2",
        x86_64="lib64/ld-linux-x86-64.so.2",
        s390x="lib/ld64.so.1",
    )

    _32BIT_USERSPACE_ARCHITECTURE = dict(
        aarch64="armv7l",
        armv8l="armv7l",
        ppc64le="ppc",
        x86_64="i686",
        i686="i686",
        armv7l="armv7l",
    )

    _WINDOWS_TRANSLATIONS = dict(AMD64="x86_64")

    def _get_platform_architecture(self):
        architecture = platform.machine()

        # Translate the windows architectures we know of to architectures
        # we can work with.
        if sys.platform == "win32":
            architecture = self._WINDOWS_TRANSLATIONS[architecture]

        if platform.architecture()[0] == "32bit":
            userspace = self._32BIT_USERSPACE_ARCHITECTURE[architecture]
            if userspace:
                architecture = userspace

        return architecture

    def __init__(self, *, machine=None):
        super().__init__()
        if machine is None:
            self._machine = self._get_platform_architecture()
        else:
            self._machine = machine

    def _setUp(self):
        super()._setUp()

        patcher = mock.patch("platform.machine")
        self.mock_machine = patcher.start()
        self.mock_machine.return_value = self._machine
        self.addCleanup(patcher.stop)

        self.core_path = self.useFixture(fixtures.TempDir()).path
        patcher = mock.patch("snapcraft.internal.common.get_installed_snap_path")
        mock_core_path = patcher.start()
        mock_core_path.return_value = self.core_path
        self.addCleanup(patcher.stop)

        self.content_dirs = set([])
        mock_content_dirs = fixtures.MockPatch(
            "snapcraft.project._project.Project._get_provider_content_dirs",
            return_value=self.content_dirs,
        )
        self.useFixture(mock_content_dirs)

        # Create file to represent the linker so it is found
        linker_path = os.path.join(self.core_path, self._LINKER_FOR_ARCH[self._machine])
        os.makedirs(os.path.dirname(linker_path), exist_ok=True)
        real_linker = os.path.join(self.core_path, "usr", "lib", "ld-2.23.so")
        os.makedirs(os.path.dirname(real_linker), exist_ok=True)
        open(real_linker, "w").close()
        os.symlink(
            os.path.relpath(real_linker, os.path.dirname(linker_path)), linker_path
        )


class FakeSnapcraftIsASnap(fixtures.Fixture):
    def _setUp(self):
        super()._setUp()

        self.useFixture(fixtures.EnvironmentVariable("SNAP", "/snap/snapcraft/current"))
        self.useFixture(fixtures.EnvironmentVariable("SNAP_NAME", "snapcraft"))
        self.useFixture(fixtures.EnvironmentVariable("SNAP_VERSION", "4.0"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict

from testtools.matchers import Equals, Is

from snapcraft.internal.meta import errors
from snapcraft.internal.meta.plugs import ContentPlug, Plug
from tests import unit


class GenericPlugTests(unit.TestCase):
    def test_plug_name(self):
        plug_name = "plug-test"

        plug = Plug(plug_name=plug_name)
        plug_from_dict = Plug.from_dict(
            plug_dict={"interface": "somevalue"}, plug_name=plug_name
        )

        self.assertEqual(plug_name, plug.plug_name)
        self.assertEqual(plug_name, plug_from_dict.plug_name)

    def test_from_empty_dict(self):
        plug_dict = OrderedDict({})
        plug_name = "plug-test"

        plug = Plug.from_dict(plug_dict=plug_dict, plug_name=plug_name)

        plug.validate()

    def test_from_valid_dict(self):
        plug_dict = OrderedDict({"interface": "somevalue", "someprop": "somevalue"})
        plug_name = "plug-test"

        plug = Plug.from_dict(plug_dict=plug_dict, plug_name=plug_name)

        plug.validate()

    def test_from_object_none(self):
        plug = Plug.from_object(plug_name="plug-name", plug_object=None)
        plug.validate()

        self.assertThat(plug._plug_dict, Equals(dict()))

    def test_from_object_string(self):
        plug = Plug.from_object(plug_name="plug-name", plug_object="some-interface")
        plug.validate()

        self.assertThat(plug._plug_dict["interface"], Equals("some-interface"))
        self.assertThat(plug.use_string_representation, Equals(True))

    def test_from_object_dict(self):
        plug_dict = OrderedDict(
            {"interface": "some-interface", "someprop": "somevalue"}
        )
        plug_name = "plug-test"

        plug = Plug.from_object(plug_object=plug_dict, plug_name=plug_name)

        plug.validate()
        self.assertThat(plug._plug_dict["interface"], Equals("some-interface"))


class ContentPlugTests(unit.TestCase):
    def test_invalid_target_raises_exception(self):
        plug = ContentPlug(plug_name="plug-test", target="")

        self.assertRaises(errors.PlugValidationError, plug.validate)

    def test_invalid_target_from_dict_raise_exception(self):
        plug_dict = OrderedDict({"interface": "content", "target": ""})

        plug = Plug.from_dict(plug_dict=plug_dict, plug_name="plug-test")

        self.assertRaises(errors.PlugValidationError, plug.validate)

    def test_content_defaults_to_name(self):
        plug = ContentPlug(plug_name="plug-test", target="target")

        self.assertEqual("plug-test", plug.content)

    def test_basic_from_dict(self):
        plug_dict = OrderedDict(
            {
                "interface": "content",
                "content": "content",
                "target": "target",
                "default-provider": "gtk-common-themes:gtk-3-themes",
            }
        )
        plug_name = "plug-test"
        plug_provider = "gtk-common-themes"

        plug = ContentPlug.from_dict(plug_dict=plug_dict, plug_name=plug_name)
        plug.validate()

        # default-provider is not written as <provider>:<name>, just <provider>.
        transformed_dict = plug_dict.copy()
        transformed_dict["default-provider"] = plug_provider

        self.assertEqual(transformed_dict, plug.to_yaml_object())
        self.assertEqual(plug_name, plug.plug_name)
        self.assertEqual(plug_dict["content"], plug.content)
        self.assertEqual(plug_dict["target"], plug.target)
        self.assertEqual(plug_provider, plug.provider)

    def test_basic_from_dict_no_default(self):
        plug_dict = OrderedDict(
            {"interface": "content", "content": "content", "target": "target"}
        )
        plug_name = "plug-test"
        plug = ContentPlug.from_dict(plug_dict=plug_dict, plug_name=plug_name)

        self.assertThat(plug.provider, Is(None))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import importlib
import sys
from textwrap import dedent

import click

import snapcraft
from snapcraft.internal import errors, sources
from snapcraft.project import errors as project_errors

from . import echo
from ._options import get_project

_TOPICS = {"sources": sources, "plugins": snapcraft}


@click.group()
def helpcli():
    """Help commands"""
    pass


@helpcli.command("help")
@click.argument("topic", metavar="<topic>", required=False)
@click.option(
    "--devel", is_flag=True, help="Show more details for snapcraft developers"
)
@click.option(
    "--base",
    help="Show help for specific base",
    type=click.Choice(["core", "core16", "core18", "core20"]),
)
@click.pass_context
def help_command(ctx, topic, devel, base):
    """Obtain help for a certain topic, plugin or command.

    The <topic> can either be a plugin name or one of:

    \b
        - topics
        - plugins
        - sources

    \b
    Examples:
        snapcraft help topics
        snapcraft help plugins
        snapcraft help sources
        snapcraft help go
        snapcraft help go --base core18
        snapcraft help build
    """
    if not topic:
        click.echo(ctx.parent.get_help())
        click.echo(
            dedent(
                """\

            For more help, use:
                snapcraft help topics
                snapcraft help <topic>
                snapcraft help <plugin-name>
                snapcraft help <command-name>
        """
            )
        )
    elif topic in ctx.parent.command.commands:
        ctx.info_name = topic
        click.echo(ctx.parent.command.commands[topic].get_help(ctx))
    elif topic == "topics":
        for key in _TOPICS:
            click.echo(key)
    elif topic in _TOPICS:
        _topic_help(topic, devel)
    else:
        try:
            _module_help(topic, devel, base)
        except ImportError:
            # 10 is the limit which determines ellipsis is needed
            if len(topic) > 10:
                topic = "{}...".format(topic[:10])
            echo.wrapped(
                dedent(
                    """\
    There is no help topic, plugin or command {!r}. Try:

    For topics:

        snapcraft help topics

    For valid plugins:

        snapcraft list-plugins

    Or for general help:

        snapcraft help
    """
                ).format(topic)
            )
            sys.exit(1)


def _topic_help(module_name, devel):
    if devel:
        help(_TOPICS[module_name])
    else:
        click.echo(_TOPICS[module_name].__doc__)


def _module_help(plugin_name: str, devel: bool, base: str):
    module_name = plugin_name.replace("-", "_")

    if base is None:
        try:
            base = get_project()._snap_meta.get_build_base()
        except (errors.ProjectNotFoundError, project_errors.MissingSnapcraftYamlError):
            base = "core20"

    if base == "core20":
        plugin_version = "v2"
    else:
        plugin_version = "v1"

    module = importlib.import_module(
        f"snapcraft.plugins.{plugin_version}.{module_name}"
    )
    if module.__doc__ and devel:
        help(module)
    elif module.__doc__:
        click.echo_via_pager(
            f"Displaying help for the {plugin_name!r} plugin for {base!r}.\n\n"
            + module.__doc__
        )
    else:
        click.echo("The plugin has no documentation")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict

from testtools.matchers import Equals

from snapcraft.internal.meta import errors
from snapcraft.internal.meta.slots import ContentSlot, DbusSlot, Slot
from tests import unit


class GenericSlotTests(unit.TestCase):
    def test_slot_name(self):
        slot_name = "slot-test"

        slot = Slot(slot_name=slot_name)
        slot_from_dict = Slot.from_dict(
            slot_dict={"interface": "somevalue"}, slot_name=slot_name
        )

        self.assertEqual(slot_name, slot.slot_name)
        self.assertEqual(slot_name, slot_from_dict.slot_name)

    def test_from_empty_dict(self):
        slot_dict = OrderedDict({})
        slot_name = "slot-test"

        slot = Slot.from_dict(slot_dict=slot_dict, slot_name=slot_name)

        slot.validate()

    def test_from_valid_dict(self):
        slot_dict = OrderedDict({"interface": "somevalue", "someprop": "somevalue"})
        slot_name = "slot-test"

        slot = Slot.from_dict(slot_dict=slot_dict, slot_name=slot_name)

        slot.validate()

    def test_from_object_none(self):
        slot = Slot.from_object(slot_name="slot-name", slot_object=None)
        slot.validate()

        self.assertThat(slot._slot_dict, Equals(dict()))

    def test_from_object_string(self):
        slot = Slot.from_object(slot_name="slot-name", slot_object="some-interface")
        slot.validate()

        self.assertThat(slot._slot_dict["interface"], Equals("some-interface"))
        self.assertThat(slot.use_string_representation, Equals(True))

    def test_from_object_dict(self):
        slot_dict = OrderedDict(
            {"interface": "some-interface", "someprop": "somevalue"}
        )
        slot_name = "slot-test"

        slot = Slot.from_object(slot_object=slot_dict, slot_name=slot_name)

        slot.validate()

        self.assertThat(slot._slot_dict["interface"], Equals("some-interface"))
        self.assertThat(slot._slot_dict["someprop"], Equals("somevalue"))


class ContentSlotTests(unit.TestCase):
    def test_empty(self):
        slot_dict = OrderedDict({"interface": "content"})
        slot_name = "slot-test"

        slot = ContentSlot(slot_name=slot_name)

        # "use_source_key" is default, account for that.
        transformed_dict = slot_dict.copy()
        transformed_dict["source"] = {}

        self.assertEqual(transformed_dict, slot.to_yaml_object())
        self.assertEqual(slot_name, slot.slot_name)
        self.assertRaises(errors.SlotValidationError, slot.validate)
        self.assertEqual(set(), slot.get_content_dirs(installed_path=""))

    def test_empty_from_dict(self):
        slot_dict = OrderedDict({"interface": "content"})
        slot_name = "slot-name"

        slot = Slot.from_dict(slot_dict=slot_dict, slot_name=slot_name)

        self.assertIsInstance(slot, ContentSlot)
        self.assertEqual(slot_dict, slot.to_yaml_object())
        self.assertEqual(slot_name, slot.slot_name)
        self.assertRaises(errors.SlotValidationError, slot.validate)
        self.assertEqual(set(), slot.get_content_dirs(installed_path=""))

    def test_empty_force_no_source(self):
        slot_dict = OrderedDict({"interface": "content"})
        slot_name = "slot-test"

        slot = ContentSlot(use_source_key=False, slot_name=slot_name)

        self.assertEqual(slot_dict, slot.to_yaml_object())
        self.assertEqual(slot_name, slot.slot_name)
        self.assertRaises(errors.SlotValidationError, slot.validate)
        self.assertEqual(set(), slot.get_content_dirs(installed_path=""))

    def test_explicit_content(self):
        slot_dict = OrderedDict({"interface": "content", "content": "content-test"})
        slot_name = "slot-test"

        slot = ContentSlot.from_dict(slot_dict=slot_dict, slot_name=slot_name)

        assert slot_dict == slot.to_yaml_object()

    def test_explicit_content_with_source(self):
        slot_dict = OrderedDict(
            {
                "interface": "content",
                "content": "content-test",
                "source": {"read": "/"},
            }
        )
        slot_name = "slot-test"

        slot = ContentSlot.from_dict(slot_dict=slot_dict, slot_name=slot_name)

        assert slot_dict == slot.to_yaml_object()

    def test_read_from_dict(self):
        slot_dict = OrderedDict(
            {
                "interface": "content",
                "content": "explicit-content",
                "read": ["some/path"],
            }
        )
        slot_name = "slot-test"

        slot = ContentSlot.from_dict(slot_dict=slot_dict, slot_name=slot_name)
        slot.validate()

        self.assertEqual(slot_dict, slot.to_yaml_object())
        self.assertEqual(slot_name, slot.slot_name)
        self.assertEqual(slot_dict["read"], slot.read)
        self.assertEqual(
            set(slot_dict["read"]), slot.get_content_dirs(installed_path="")
        )
        self.assertEqual(slot_dict["content"], slot.content)

    def test_read_from_dict_force_source_key(self):
        slot_dict = OrderedDict({"interface": "content", "read": ["some/path"]})
        slot_name = "slot-test"

        slot = ContentSlot.from_dict(slot_dict=slot_dict, slot_name=slot_name)
        slot.use_source_key = True
        slot.validate()
        transformed_dict = slot_dict.copy()
        transformed_dict["source"] = dict()
        transformed_dict["source"]["read"] = transformed_dict.pop("read")

        self.assertEqual(transformed_dict, slot.to_yaml_object())
        self.assertEqual(slot_name, slot.slot_name)
        self.assertEqual(slot_dict["read"], slot.read)
        self.assertEqual(
            set(slot_dict["read"]), slot.get_content_dirs(installed_path="")
        )

    def test_source_read_from_dict(self):
        slot_dict = OrderedDict(
            {"interface": "content", "source": {"read": ["some/path"]}}
        )
        slot_name = "slot-test"

        slot = ContentSlot.from_dict(slot_dict=slot_dict, slot_name=slot_name)
        slot.validate()

        self.assertEqual(slot_dict, slot.to_yaml_object())
        self.assertEqual(slot_name, slot.slot_name)
        self.assertEqual(slot_dict["source"]["read"], slot.read)
        self.assertEqual(
            set(slot_dict["source"]["read"]), slot.get_content_dirs(installed_path="")
        )

    def test_write_from_dict(self):
        slot_dict = OrderedDict({"interface": "content", "write": ["some/path"]})
        slot_name = "slot-test"

        slot = ContentSlot.from_dict(slot_dict=slot_dict, slot_name=slot_name)
        slot.validate()

        self.assertEqual(slot_dict, slot.to_yaml_object())
        self.assertEqual(slot_name, slot.slot_name)
        self.assertEqual(slot_dict["write"], slot.write)
        self.assertEqual(
            set(slot_dict["write"]), slot.get_content_dirs(installed_path="")
        )

    def test_write_from_dict_force_source_key(self):
        slot_dict = OrderedDict({"interface": "content", "write": ["some/path"]})
        slot_name = "slot-test"

        slot = ContentSlot.from_dict(slot_dict=slot_dict, slot_name=slot_name)
        slot.use_source_key = True
        slot.validate()
        transformed_dict = slot_dict.copy()
        transformed_dict["source"] = dict()
        transformed_dict["source"]["write"] = transformed_dict.pop("write")

        self.assertEqual(transformed_dict, slot.to_yaml_object())
        self.assertEqual(slot_name, slot.slot_name)
        self.assertEqual(slot_dict["write"], slot.write)
        self.assertEqual(
            set(slot_dict["write"]), slot.get_content_dirs(installed_path="")
        )

    def test_source_write_from_dict(self):
        slot_dict = OrderedDict(
            {"interface": "content", "source": {"write": ["some/path"]}}
        )
        slot_name = "slot-test"

        slot = ContentSlot.from_dict(slot_dict=slot_dict, slot_name=slot_name)
        slot.validate()

        self.assertEqual(slot_dict, slot.to_yaml_object())
        self.assertEqual(slot_name, slot.slot_name)
        self.assertEqual(slot_dict["source"]["write"], slot.write)
        self.assertEqual(
            set(slot_dict["source"]["write"]), slot.get_content_dirs(installed_path="")
        )


class DbusSlotTests(unit.TestCase):
    def test_invalid_target_raises_exception(self):
        slot = DbusSlot(slot_name="slot-test", bus="", name="")

        self.assertRaises(errors.SlotValidationError, slot.validate)

    def test_invalid_interface_from_dict_raises_exception(self):
        slot_dict = OrderedDict({"interface": "content", "bus": "", "name": ""})

        self.assertRaises(
            errors.SlotValidationError,
            DbusSlot.from_dict,
            slot_dict=slot_dict,
            slot_name="slot-test",
        )

    def test_invalid_target_from_dict_raises_exception(self):
        slot_dict = OrderedDict({"interface": "dbus", "bus": "", "name": ""})

        slot = DbusSlot.from_dict(slot_dict=slot_dict, slot_name="slot-test")

        self.assertRaises(errors.SlotValidationError, slot.validate)

    def test_basic(self):
        slot_dict = OrderedDict({"interface": "dbus", "bus": "bus", "name": "name"})
        slot_name = "slot-test"

        slot = DbusSlot(
            slot_name=slot_name, bus=slot_dict["bus"], name=slot_dict["name"]
        )
        slot.validate()

        self.assertEqual(slot_dict, slot.to_yaml_object())
        self.assertEqual(slot_name, slot.slot_name)
        self.assertEqual(slot_dict["bus"], slot.bus)
        self.assertEqual(slot_dict["name"], slot.name)

    def test_basic_from_dict(self):
        slot_dict = OrderedDict({"interface": "dbus", "bus": "bus", "name": "name"})
        slot_name = "slot-test"

        slot = DbusSlot.from_dict(slot_dict=slot_dict, slot_name=slot_name)
        slot.validate()

        self.assertEqual(slot_dict, slot.to_yaml_object())
        self.assertEqual(slot_name, slot.slot_name)
        self.assertEqual(slot_dict["bus"], slot.bus)
        self.assertEqual(slot_dict["name"], slot.name)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import contextlib
import logging
import os
import shutil
from typing import TYPE_CHECKING, Optional

from snapcraft import formatting_utils
from snapcraft.internal import errors, mountinfo, project_loader, steps

logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from snapcraft.project import Project


def _clean_part(part_name, step, config, staged_state, primed_state):
    if step.next_step() is None:
        template = "Cleaning {step} step for {part}"
    else:
        template = "Cleaning {step} step (and all subsequent steps) for {part}"

    logger.info(template.format(step=step.name, part=part_name))
    config.parts.clean_part(part_name, staged_state, primed_state, step)

    # If we just cleaned the root of a dependency tree. Return all dirty
    # dependents.
    return get_dirty_reverse_dependencies(part_name, step, config)


def get_dirty_reverse_dependencies(part_name, step, config):
    # If other parts are depending on this step of the part, they're now
    # dirty
    dirty_reverse_dependencies = set()
    reverse_dependencies = config.parts.get_reverse_dependencies(
        part_name, recursive=True
    )
    dirty_step = steps.dirty_step_if_dependency_changes(step)
    for reverse_dependency in reverse_dependencies:
        if not reverse_dependency.should_step_run(dirty_step):
            dirty_reverse_dependencies.add(reverse_dependency)

    return dirty_reverse_dependencies


def _clean_parts(part_names, step, config, staged_state, primed_state):
    if not step:
        step = steps.next_step(None)

    for part_name in part_names:
        dirty_parts = _clean_part(part_name, step, config, staged_state, primed_state)
        dirty_part_names = {p.name for p in dirty_parts}

        parts_not_being_cleaned = dirty_part_names.difference(part_names)
        if parts_not_being_cleaned:
            logger.warning(
                "Cleaned {!r}, which makes the following {} out of date: "
                "{}".format(
                    part_name,
                    formatting_utils.pluralize(
                        parts_not_being_cleaned, "part", "parts"
                    ),
                    formatting_utils.humanize_list(parts_not_being_cleaned, "and"),
                )
            )


def _remove_directory(directory: str) -> None:
    if os.path.isdir(directory):
        shutil.rmtree(directory)


def _remove_directory_if_empty(directory):
    if os.path.isdir(directory) and not os.listdir(directory):
        os.rmdir(directory)


def _cleanup_common_directories(config, project: "Project") -> None:
    max_step = None
    for part in config.all_parts:
        with contextlib.suppress(errors.NoLatestStepError):
            step = part.latest_step()
            if not max_step or step > max_step:
                max_step = step

    next_step = steps.next_step(max_step)
    if next_step:
        _cleanup_common_directories_for_step(next_step, project)


def _cleanup_common_directories_for_step(step, project: "Project", parts=None):
    if not parts:
        parts = []

    being_tried = False
    message: Optional[str] = None
    if step <= steps.PRIME:
        # Remove the priming area. Only remove the actual 'prime' directory if
        # it's NOT being used in 'snap try'. We'll know that if it's
        # bind-mounted somewhere.
        mounts = mountinfo.MountInfo()
        try:
            mounts.for_root(project.prime_dir)
        except errors.RootNotMountedError:
            remove_dir = True
            message = "Cleaning up priming area"
        else:
            remove_dir = False
            message = (
                "Cleaning up priming area, but not removing as it's in "
                "use by 'snap try'"
            )
            being_tried = True
        finally:
            # It does not matter if prime is being tried inside the managed-host
            # or not, we cannot delete prime as it is really a mount from the
            # outside host.
            if project._is_managed_host:
                remove_dir = False
                message = None
        _cleanup_common(
            project.prime_dir, steps.PRIME, message, parts, remove_dir=remove_dir
        )

    if step <= steps.STAGE:
        # Remove the staging area.
        _cleanup_common(
            project.stage_dir, steps.STAGE, "Cleaning up staging area", parts
        )

    if step <= steps.PULL:
        # Remove the parts directory (but leave local plugins alone).
        _cleanup_parts_dir(project.parts_dir, project.local_plugins_dir, parts)

    if not being_tried and not project._is_managed_host:
        _remove_directory_if_empty(project.prime_dir)
    _remove_directory_if_empty(project.stage_dir)
    _remove_directory_if_empty(project.parts_dir)


def _cleanup_common(directory, step, message, parts, *, remove_dir=True):
    if os.path.isdir(directory):
        if message is not None:
            logger.info(message)
        if remove_dir:
            shutil.rmtree(directory)
        else:
            # Don't delete the parent directory, but delete its contents
            for f in os.scandir(directory):
                if f.is_dir(follow_symlinks=False):
                    shutil.rmtree(f.path)
                elif f.is_file(follow_symlinks=False):
                    os.remove(f.path)
    for part in parts:
        part.mark_cleaned(step)


def _cleanup_parts_dir(parts_dir, local_plugins_dir, parts):
    if os.path.exists(parts_dir):
        logger.info("Cleaning up parts directory")
        for subdirectory in os.listdir(parts_dir):
            path = os.path.join(parts_dir, subdirectory)
            if path != local_plugins_dir:
                try:
                    shutil.rmtree(path)
                except NotADirectoryError:
                    os.remove(path)
    for part in parts:
        part.mark_cleaned(steps.BUILD)
        part.mark_cleaned(steps.PULL)


def clean(project: "Project", parts, step=None):
    # step defaults to None because that's how it comes from docopt when it's
    # not set.
    if not step:
        step = steps.PULL

    if not parts and step == steps.PULL:
        _cleanup_common_directories_for_step(step, project)
        return

    config = project_loader.load_config(project)

    if not parts and step <= steps.PRIME:
        # If we've been asked to clean stage or prime without being given
        # specific parts, just blow away those directories instead of
        # doing it per part (it would just be a waste of time).
        _cleanup_common_directories_for_step(step, project, parts=config.all_parts)

        # No need to continue if that's all that was required
        if step >= steps.STAGE:
            return

    if parts:
        config.parts.validate(parts)
    else:
        parts = [part.name for part in config.all_parts]

    staged_state = config.get_project_state(steps.STAGE)
    primed_state = config.get_project_state(steps.PRIME)

    _clean_parts(parts, step, config, staged_state, primed_state)

    _cleanup_common_directories(config, project)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess
from textwrap import dedent

from testtools.matchers import Contains

from tests import integration
from tests.integration import repo


class BuildPackageGrammarTestCase(integration.TestCase):
    def setUp(self):
        super().setUp()
        if self._hello_is_installed():
            self.fail(
                "This integration test cannot run if you already have the "
                "'hello' package installed. Please uninstall it before "
                "running this test."
            )

    def tearDown(self):
        super().tearDown()

        # Remove hello. This is safe since the test fails if hello was already
        # installed.
        try:
            subprocess.check_output(
                ["sudo", "apt", "remove", "hello", "-y"], stderr=subprocess.STDOUT
            )
        except subprocess.CalledProcessError as e:
            self.fail("unable to remove 'hello': {}".format(e.output))

    def _hello_is_installed(self):
        return repo.is_package_installed("hello")

    def test_simple(self):
        """Test that grammar installs standalone build package."""

        self.run_snapcraft(["pull"], "build-package-grammar")

        self.assertTrue(self._hello_is_installed())

    def test_try(self):
        """Test that 'try' installs valid build packages."""

        self.run_snapcraft(["pull"], "build-package-grammar-try")

        self.assertTrue(self._hello_is_installed())

    def test_try_skipped(self):
        """Test that 'try' skips invalid build packages."""

        self.run_snapcraft(["pull"], "build-package-grammar-try-skip")

        self.assertFalse(self._hello_is_installed())

    def test_try_else(self):
        """Test that 'try' moves to the 'else' branch if body is invalid"""

        self.run_snapcraft(["pull"], "build-package-grammar-try-else")

        self.assertTrue(self._hello_is_installed())

    def test_on_other_arch(self):
        """Test that 'on' fetches nothing when running on another arch."""

        self.run_snapcraft(["pull"], "build-package-grammar-on")

        self.assertFalse(self._hello_is_installed())

    def test_on_other_arch_else(self):
        """Test that 'on' moves to the 'else' branch if on other arch."""

        self.run_snapcraft(["pull"], "build-package-grammar-on-else")

        self.assertTrue(self._hello_is_installed())

    def test_on_other_arch_else_fail(self):
        """Test that 'on' fails with an error if it hits an 'else fail'."""

        exception = self.assertRaises(
            subprocess.CalledProcessError,
            self.run_snapcraft,
            ["pull"],
            "build-package-grammar-fail",
        )

        self.assertThat(
            exception.output,
            Contains("Unable to satisfy 'on other-arch', failure forced"),
        )

    def test_global_build_package_on_other_arch_else(self):
        """Test that grammar works in global build packages as well."""

        self.run_snapcraft(["pull"], "build-package-grammar-global")

        self.assertTrue(self._hello_is_installed())

    def test_to_other_arch(self):
        """Test that 'to' fetches nothing when building for another arch."""

        self.construct_yaml(
            parts=dedent(
                """\
            my-part:
              plugin: nil
              build-packages:
              - to other-arch:
                - hello
            """
            )
        )
        self.run_snapcraft(["pull"])
        self.assertFalse(self._hello_is_installed())

    def test_to_other_arch_else(self):
        """Test that 'to' moves to the 'else' branch if on other arch."""

        self.construct_yaml(
            parts=dedent(
                """\
            my-part:
              plugin: nil
              build-packages:
              - to other-arch:
                - foo
              - else:
                - hello
            """
            )
        )
        self.run_snapcraft(["pull"])
        self.assertTrue(self._hello_is_installed())

    def test_to_other_arch_else_fail(self):
        """Test that 'on' fails with an error if it hits an 'else fail'."""

        self.construct_yaml(
            parts=dedent(
                """\
            my-part:
              plugin: nil
              build-packages:
              - to other-arch:
                - foo
              - else fail
            """
            )
        )
        self.assertThat(
            self.assertRaises(
                subprocess.CalledProcessError, self.run_snapcraft, ["pull"]
            ).output,
            Contains("Unable to satisfy 'to other-arch', failure forced"),
        )

    def test_global_build_package_to_other_arch_else(self):
        """Test that grammar works in global build packages as well."""

        self.construct_yaml(
            build_packages=dedent(
                """\
            - to other-arch:
              - foo
            - else:
              - hello
            """
            )
        )
        self.run_snapcraft(["pull"])
        self.assertTrue(self._hello_is_installed())

    def test_on_to_other_arch(self):
        """Test that 'on to' fetches nothing when building for another arch."""

        self.construct_yaml(
            parts=dedent(
                """\
            my-part:
              plugin: nil
              build-packages:
              - on i386 to other-arch:
                - hello
            """
            )
        )
        self.run_snapcraft(["pull"])
        self.assertFalse(self._hello_is_installed())

    def test_on_to_other_arch_else(self):
        """Test that 'on to' moves to the 'else' branch if on other arch."""

        self.construct_yaml(
            parts=dedent(
                """\
            my-part:
              plugin: nil
              build-packages:
              - on i386 to other-arch:
                - foo
              - else:
                - hello
            """
            )
        )
        self.run_snapcraft(["pull"])
        self.assertTrue(self._hello_is_installed())

    def test_on_to_other_arch_else_fail_on(self):
        self.construct_yaml(
            parts=dedent(
                """\
            my-part:
              plugin: nil
              build-packages:
              - on i386 to other-arch:
                - foo
              - else fail
            """
            )
        )
        self.assertThat(
            self.assertRaises(
                subprocess.CalledProcessError, self.run_snapcraft, ["pull"]
            ).output,
            Contains("Unable to satisfy 'on i386 to other-arch', failure forced"),
        )

    def test_on_to_other_arch_else_fail_to(self):
        self.construct_yaml(
            parts=dedent(
                """\
            my-part:
              plugin: nil
              build-packages:
              - on amd64 to other-arch:
                - foo
              - else fail
            """
            )
        )
        self.assertThat(
            self.assertRaises(
                subprocess.CalledProcessError, self.run_snapcraft, ["pull"]
            ).output,
            Contains("Unable to satisfy 'on amd64 to other-arch', failure forced"),
        )

    def test_global_build_package_on_to_other_arch_else(self):
        """Test that on..to works in global build packages as well."""

        self.construct_yaml(
            build_packages=dedent(
                """\
            - on i386 to other-arch:
              - foo
            - else:
              - hello
            """
            )
        )
        self.run_snapcraft(["pull"])
        self.assertTrue(self._hello_is_installed())



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import snapcraft.yaml_utils.errors
from snapcraft.project import Project as _Project
from tests import unit


class ProjectBaseTest(unit.TestCase):
    def make_snapcraft_project(self, snapcraft_yaml, project_kwargs=None):
        snapcraft_yaml_file_path = self.make_snapcraft_yaml(snapcraft_yaml)
        if project_kwargs is None:
            project_kwargs = dict()
        return _Project(
            snapcraft_yaml_file_path=snapcraft_yaml_file_path, **project_kwargs
        )

    def assertValidationPasses(self, snapcraft_yaml):
        project = self.make_snapcraft_project(snapcraft_yaml)
        project.info.validate_raw_snapcraft()

        return project.info.get_raw_snapcraft()

    def assertValidationRaises(self, snapcraft_yaml):
        project = self.make_snapcraft_project(snapcraft_yaml)

        return self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError,
            project.info.validate_raw_snapcraft,
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Type

_VALID_RISKS = ["stable", "candidate", "beta", "edge"]


class Channel:
    def __repr__(self) -> str:
        return "{!r}".format(self._channel)

    def __str__(self) -> str:
        return self._channel

    def __eq__(self, other) -> bool:
        return (
            self.track == other.track
            and self.risk == other.risk
            and self.branch == other.branch
        )

    @classmethod
    def from_channel_tuple(
        cls: Type["Channel"], *, track: str, risk: str, branch: Optional[str]
    ) -> "Channel":
        if track and risk and branch:
            channel = "{}/{}/{}".format(track, risk, branch)
        elif track and risk:
            channel = "{}/{}".format(track, risk)
        elif risk and branch:
            channel = "{}/{}".format(risk, branch)
        elif risk:
            channel = risk
        else:
            raise RuntimeError(
                "Incorrect channel tuple {}/{}/{}.".format(track, risk, branch)
            )

        return cls(channel)

    def __init__(self, channel: str) -> None:
        channel_parts = channel.split("/")
        if len(channel_parts) == 1 and channel_parts[0] in _VALID_RISKS:
            self._track = None
            self._risk = channel_parts[0]
            self._branch = None
        elif len(channel_parts) == 3 and channel_parts[1] in _VALID_RISKS:
            self._track = channel_parts[0]
            self._risk = channel_parts[1]
            self._branch = channel_parts[2]
        elif len(channel_parts) == 2 and channel_parts[0] in _VALID_RISKS:
            self._track = None
            self._risk = channel_parts[0]
            self._branch = channel_parts[1]
        elif len(channel_parts) == 2 and channel_parts[1] in _VALID_RISKS:
            self._track = channel_parts[0]
            self._risk = channel_parts[1]
            self._branch = None
        else:
            raise RuntimeError("Channel logic failed for: {!r}".format(channel))
        self._channel = channel

    @property
    def track(self) -> str:
        if self._track is None:
            return "latest"
        else:
            return self._track

    @property
    def risk(self) -> str:
        return self._risk

    @property
    def branch(self) -> Optional[str]:
        return self._branch



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from typing import TYPE_CHECKING, Dict, Type, Union

from snapcraft.internal import errors

from . import v1, v2

PluginTypes = Union[Type[v1.PluginV1], Type[v2.PluginV2]]

# TODO: segregate into base specific plugins.
if sys.platform == "linux" or TYPE_CHECKING:
    _PLUGINS: Dict[str, Dict[str, PluginTypes]] = {
        "v1": {
            "ant": v1.AntPlugin,
            "autotools": v1.AutotoolsPlugin,
            "catkin": v1.CatkinPlugin,
            "catkin-tools": v1.CatkinToolsPlugin,
            "cmake": v1.CMakePlugin,
            "colcon": v1.ColconPlugin,
            "conda": v1.CondaPlugin,
            "crystal": v1.CrystalPlugin,
            "dotnet": v1.DotNetPlugin,
            "dump": v1.DumpPlugin,
            "flutter": v1.FlutterPlugin,
            "go": v1.GoPlugin,
            "godeps": v1.GodepsPlugin,
            "gradle": v1.GradlePlugin,
            "kbuild": v1.KBuildPlugin,
            "kernel": v1.KernelPlugin,
            "make": v1.MakePlugin,
            "maven": v1.MavenPlugin,
            "meson": v1.MesonPlugin,
            "nil": v1.NilPlugin,
            "nodejs": v1.NodePlugin,
            "plainbox-provider": v1.PlainboxProviderPlugin,
            "python": v1.PythonPlugin,
            "qmake": v1.QmakePlugin,
            "ruby": v1.RubyPlugin,
            "rust": v1.RustPlugin,
            "scons": v1.SconsPlugin,
            "waf": v1.WafPlugin,
        },
        "v2": {
            "autotools": v2.AutotoolsPlugin,
            "catkin": v2.CatkinPlugin,
            "catkin-tools": v2.CatkinToolsPlugin,
            "cmake": v2.CMakePlugin,
            "colcon": v2.ColconPlugin,
            "conda": v2.CondaPlugin,
            "dump": v2.DumpPlugin,
            "go": v2.GoPlugin,
            "make": v2.MakePlugin,
            "meson": v2.MesonPlugin,
            "nil": v2.NilPlugin,
            "npm": v2.NpmPlugin,
            "python": v2.PythonPlugin,
            "qmake": v2.QMakePlugin,
            "rust": v2.RustPlugin,
        },
    }
else:
    # We cannot import the plugins on anything but linux.
    _PLUGINS = {"v1": {}, "v2": {}}


def get_plugin_for_base(plugin_name: str, *, build_base: str) -> PluginTypes:
    """
    Return a suitable plugin implementation for build_base.

    :raises errors.PluginError: when a plugin implementation cannot be found
                                for build_base.
    """
    # TODO: segregate this more and remove the checks UnsupportedBase checks
    # from the plugins

    # Default for unknown, and core20
    plugin_version = "v2"
    # Others use v1
    if build_base in ("core", "core16", "core18"):
        plugin_version = "v1"
    elif build_base == "":
        # This should never happen when using build_base from Project.
        raise RuntimeError("'build_base' cannot be unset.")
        plugin_version = "v1"

    try:
        plugin_classes = _PLUGINS[plugin_version]
        return plugin_classes[plugin_name]
    except KeyError:
        raise errors.PluginError(f"unknown plugin: {plugin_name!r}")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest import mock

from testtools.matchers import Contains, Equals, HasLength, Not

from snapcraft.internal import errors
from snapcraft.plugins.v1 import godeps
from tests import unit

from . import PluginsV1BaseTestCase


class GodepsPluginBaseTest(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        patcher = mock.patch("snapcraft.internal.common.run")
        self.run_mock = patcher.start()
        self.addCleanup(patcher.stop)

        class Options:
            source = "src"
            go_channel = "1.15/stable"
            go_importpath = "github.com/foo/bar"
            godeps_file = "dependencies.tsv"
            go_packages = []

        self.options = Options()


class GoPluginPropertiesTest(unit.TestCase):
    def test_schema(self):
        schema = godeps.GodepsPlugin.schema()

        properties = schema["properties"]
        for expected in ["godeps-file", "go-importpath"]:
            self.assertTrue(
                expected in properties,
                "Expected {!r} to be included in properties".format(expected),
            )

        # Check go-channel
        go_channel = properties["go-channel"]
        for expected in ["type", "default"]:
            self.assertTrue(
                expected in go_channel,
                "Expected {!r} to be included in 'go-channel'".format(expected),
            )

        go_channel_type = go_channel["type"]
        self.assertThat(
            go_channel_type,
            Equals("string"),
            'Expected "go-channel" "type" to be "string", but '
            'it was "{}"'.format(go_channel_type),
        )

        go_channel_default = go_channel["default"]
        self.assertThat(
            go_channel_default,
            Equals("1.15/stable"),
            'Expected "go-channel" "default" to be '
            '"1.15/stable", but it was "{}"'.format(go_channel_default),
        )

        # Check godeps-file
        godeps_file = properties["godeps-file"]
        for expected in ["type", "default"]:
            self.assertTrue(
                expected in godeps_file,
                "Expected {!r} to be included in 'godeps-file'".format(expected),
            )

        godeps_file_type = godeps_file["type"]
        self.assertThat(
            godeps_file_type,
            Equals("string"),
            'Expected "godeps-file" "type" to be "string", but '
            'it was "{}"'.format(godeps_file_type),
        )

        godeps_file_default = godeps_file["default"]
        self.assertThat(
            godeps_file_default,
            Equals("dependencies.tsv"),
            'Expected "godeps-file" "default" to be '
            '"dependencies.tsv", but it was "{}"'.format(godeps_file_default),
        )

        # Check go-importpath
        go_importpath = properties["go-importpath"]
        for expected in ["type"]:
            self.assertTrue(
                expected in go_importpath,
                "Expected {!r} to be included in 'go-importpath'".format(expected),
            )

        go_importpath_type = go_importpath["type"]
        self.assertThat(
            go_importpath_type,
            Equals("string"),
            'Expected "go-importpath" "type" to be "string", but '
            'it was "{}"'.format(go_importpath_type),
        )

        # go-importpath should be required
        self.assertTrue(
            "go-importpath" in schema["required"],
            'Expeced "go-importpath" to be required',
        )

        # Check go-packages
        go_packages = properties["go-packages"]
        for expected in ["type", "default"]:
            self.assertTrue(
                expected in go_packages,
                "Expected {!r} to be included in 'go-packages'".format(expected),
            )

        go_packages_type = go_packages["type"]
        self.assertThat(
            go_packages_type,
            Equals("array"),
            'Expected "go-packages" "type" to be "array", but '
            'it was "{}"'.format(go_packages_type),
        )

        go_packages_default = go_packages["default"]
        self.assertThat(
            go_packages_default,
            Equals([]),
            'Expected "go-packages" "default" to be "[]", but '
            'it was "{}"'.format(go_packages_default),
        )

    def test_get_pull_properties(self):
        expected_pull_properties = ["go-channel", "godeps-file", "go-importpath"]
        resulting_pull_properties = godeps.GodepsPlugin.get_pull_properties()

        self.assertThat(
            resulting_pull_properties, HasLength(len(expected_pull_properties))
        )

        self.assertThat(resulting_pull_properties, Equals(expected_pull_properties))

    def test_get_build_properties(self):
        expected_build_properties = ["go-packages"]
        resulting_build_properties = godeps.GodepsPlugin.get_build_properties()

        self.assertThat(
            resulting_build_properties, HasLength(len(expected_build_properties))
        )

        self.assertThat(resulting_build_properties, Equals(expected_build_properties))


class GodepsPluginTest(GodepsPluginBaseTest):
    def test_build_environment(self):
        plugin = godeps.GodepsPlugin("test-part", self.options, self.project)

        os.makedirs(plugin.options.source)
        os.makedirs(os.path.join(plugin.installdir, "lib"))
        os.makedirs(os.path.join(plugin.installdir, "usr", "lib"))
        os.makedirs(os.path.join(plugin.project.stage_dir, "lib"))
        os.makedirs(os.path.join(plugin.project.stage_dir, "usr", "lib"))
        plugin.pull()

        self.assertThat(self.run_mock.call_count, Equals(6))
        for call_args in self.run_mock.call_args_list:
            env = call_args[1]["env"]
            self.assertTrue("GOPATH" in env, "Expected environment to include GOPATH")
            self.assertThat(env["GOPATH"], Equals(plugin._gopath))

            self.assertTrue("PATH" in env, "Expected environment to include PATH")
            self.assertTrue(
                os.path.join(plugin._gopath, "bin") in env["PATH"],
                "Expected $PATH to include $GOPATH/bin",
            )

            self.assertTrue(
                "CGO_LDFLAGS" in env, "Expected environment to include CGO_LDFLAGS"
            )
            expected_flags = [
                "-L{}/lib".format(plugin.installdir),
                "-L{}/usr/lib".format(plugin.installdir),
                "-L{}/lib".format(plugin.project.stage_dir),
                "-L{}/usr/lib".format(plugin.project.stage_dir),
            ]
            for flag in expected_flags:
                self.assertTrue(
                    flag in env["CGO_LDFLAGS"],
                    "Expected $CGO_LDFLAGS to include {!r}, but it was "
                    '"{}"'.format(flag, env["CGO_LDFLAGS"]),
                )

    def test_pull(self):
        plugin = godeps.GodepsPlugin("test-part", self.options, self.project)

        os.makedirs(plugin.options.source)

        plugin.pull()

        assert self.run_mock.mock_calls == [
            mock.call(
                ["go", "get", "-d", "github.com/rogpeppe/godeps"],
                cwd=plugin._gopath_src,
                env=mock.ANY,
            ),
            mock.call(
                ["git", "checkout", "4e9e0ee19b60b13eb79915933f44d8ed5f268bdd"],
                cwd=plugin._gopath_src + "/github.com/pelletier/go-toml",
                env=mock.ANY,
            ),
            mock.call(
                ["git", "checkout", "d6ce6262d87e3a4e153e86023ff56ae771554a41"],
                cwd=plugin._gopath_src + "/github.com/kisielk/gotool",
                env=mock.ANY,
            ),
            mock.call(
                ["git", "checkout", "1937f90a1bb43667aff4059b1bab13eb15121e8e"],
                cwd=plugin._gopath_src + "/golang.org/x/tools",
                env=mock.ANY,
            ),
            mock.call(
                ["go", "install", "github.com/rogpeppe/godeps"],
                cwd=plugin._gopath_src,
                env=mock.ANY,
            ),
            mock.call(
                [
                    "godeps",
                    "-t",
                    "-u",
                    os.path.join(plugin.sourcedir, self.options.godeps_file),
                ],
                cwd=plugin._gopath_src,
                env=mock.ANY,
            ),
        ]

        self.assertTrue(os.path.exists(plugin._gopath))
        self.assertTrue(os.path.exists(plugin._gopath_src))
        self.assertFalse(os.path.exists(plugin._gopath_pkg))
        self.assertFalse(os.path.exists(plugin._gopath_bin))

        sourcedir = os.path.join(plugin._gopath_src, "github.com", "foo", "bar")
        self.assertTrue(os.path.islink(sourcedir))
        self.assertThat(os.readlink(sourcedir), Equals(plugin.sourcedir))

    def test_clean_pull(self):
        plugin = godeps.GodepsPlugin("test-part", self.options, self.project)

        os.makedirs(plugin.options.source)

        plugin.pull()

        self.assertTrue(os.path.exists(plugin._gopath))

        plugin.clean_pull()

        self.assertFalse(os.path.exists(plugin._gopath))

    def test_build(self):
        plugin = godeps.GodepsPlugin("test-part", self.options, self.project)

        os.makedirs(plugin.options.source)

        plugin.pull()

        os.makedirs(plugin._gopath_bin)
        os.makedirs(plugin.builddir)

        self.run_mock.reset_mock()

        open(os.path.join(plugin._gopath_bin, "test-binary"), "w").close()
        open(os.path.join(plugin._gopath_bin, "godeps"), "w").close()
        plugin.build()

        self.assertThat(self.run_mock.call_count, Equals(1))
        self.run_mock.assert_has_calls(
            [
                mock.call(
                    ["go", "install", "./github.com/foo/bar/..."],
                    cwd=plugin._gopath_src,
                    env=mock.ANY,
                )
            ]
        )

        self.assertTrue(os.path.exists(plugin._gopath))
        self.assertTrue(os.path.exists(plugin._gopath_src))
        self.assertTrue(os.path.exists(plugin._gopath_bin))
        self.assertTrue(
            os.path.exists(os.path.join(plugin.installdir, "bin", "test-binary"))
        )

        # Assert that the godeps binary was NOT copied
        self.assertFalse(
            os.path.exists(os.path.join(plugin.installdir, "bin", "godeps"))
        )

    def test_build_with_go_packages(self):
        self.options.go_packages = ["github.com/foo/bar/cmd/my-command"]

        plugin = godeps.GodepsPlugin("test-part", self.options, self.project)

        os.makedirs(plugin.options.source)

        plugin.pull()

        os.makedirs(plugin._gopath_bin)
        os.makedirs(plugin.builddir)

        self.run_mock.reset_mock()

        open(os.path.join(plugin._gopath_bin, "test-binary"), "w").close()
        open(os.path.join(plugin._gopath_bin, "godeps"), "w").close()
        plugin.build()

        self.assertThat(self.run_mock.call_count, Equals(1))
        self.run_mock.assert_has_calls(
            [
                mock.call(
                    ["go", "install", "github.com/foo/bar/cmd/my-command"],
                    cwd=plugin._gopath_src,
                    env=mock.ANY,
                )
            ]
        )

        self.assertTrue(os.path.exists(plugin._gopath))
        self.assertTrue(os.path.exists(plugin._gopath_src))
        self.assertTrue(os.path.exists(plugin._gopath_bin))
        self.assertTrue(
            os.path.exists(os.path.join(plugin.installdir, "bin", "test-binary"))
        )

        # Assert that the godeps binary was NOT copied
        self.assertFalse(
            os.path.exists(os.path.join(plugin.installdir, "bin", "godeps"))
        )

    def test_clean_build(self):
        plugin = godeps.GodepsPlugin("test-part", self.options, self.project)

        os.makedirs(plugin.options.source)

        plugin.pull()

        os.makedirs(plugin._gopath_bin)
        os.makedirs(plugin._gopath_pkg)
        os.makedirs(plugin.builddir)

        plugin.build()

        self.assertTrue(os.path.exists(plugin._gopath))
        self.assertTrue(os.path.exists(plugin._gopath_src))
        self.assertTrue(os.path.exists(plugin._gopath_pkg))
        self.assertTrue(os.path.exists(plugin._gopath_bin))

        plugin.clean_build()

        self.assertTrue(os.path.exists(plugin._gopath))
        self.assertTrue(os.path.exists(plugin._gopath_src))
        self.assertFalse(os.path.exists(plugin._gopath_pkg))
        self.assertFalse(os.path.exists(plugin._gopath_bin))


class GodepsPluginToolSetupTest(GodepsPluginBaseTest):
    def test_snap(self):
        plugin = godeps.GodepsPlugin("test-part", self.options, self.project)

        self.assertThat(plugin.build_packages, Not(Contains("golang-go")))
        self.assertThat(plugin.build_snaps, Contains("go/1.15/stable"))

    def test_build_packages(self):
        self.options.go_channel = ""

        plugin = godeps.GodepsPlugin("test-part", self.options, self.project)

        self.assertThat(plugin.build_packages, Contains("golang-go"))
        self.assertThat(plugin.build_snaps, Not(Contains("go/1.15/stable")))


class GodepsPluginUnsupportedBaseTest(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        self.project._snap_meta.base = "unsupported-base"

        class Options:
            source = "dir"
            go_channel = "1.15/stable"

        self.options = Options()

    def test_unsupported_base_using_snap(self):
        self.assertRaises(
            errors.PluginBaseError,
            godeps.GodepsPlugin,
            "test-part",
            self.options,
            self.project,
        )

    def test_unsupported_base_using_without_snap_raises(self):
        self.options.go_channel = ""

        self.assertRaises(
            errors.PluginBaseError,
            godeps.GodepsPlugin,
            "test-part",
            self.options,
            self.project,
        )



#!/usr/bin/env python
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import tempfile

from launchpadlib.launchpad import Launchpad

ACTIVE_DISTROS = ("xenial", "artful", "bionic")
ACTIVE_ARCHITECTURES = ("amd64", "i386", "armhf", "arm64")


def main():
    try:
        cookie_file_path = save_cookie()
        for distro, architecture, version in snapcraft_ppa_packages():
            request_autopkgtest_execution(
                cookie_file_path, distro, architecture, version
            )
    finally:
        os.remove(cookie_file_path)


def save_cookie():
    cookie_file = tempfile.NamedTemporaryFile(delete=False)
    cookie_file.write(
        "autopkgtest.ubuntu.com\tTRUE\t/\tTRUE\t0\tsession\t{}".format(
            os.environ.get("SNAPCRAFT_AUTOPKGTEST_COOKIE")
        )
    )
    cookie_file.close()
    return cookie_file.name


def snapcraft_ppa_packages():
    launchpad = Launchpad.login_anonymously("snappy-m-o", "production")
    ubuntu = launchpad.distributions["ubuntu"]
    snapcraft_daily_ppa = launchpad.people["snappy-dev"].getPPAByName(
        name="snapcraft-daily"
    )

    for distro in ACTIVE_DISTROS:
        for architecture in ACTIVE_ARCHITECTURES:
            distro_arch = ubuntu.getSeries(name_or_version=distro).getDistroArchSeries(
                archtag=architecture
            )
            for package in snapcraft_daily_ppa.getPublishedBinaries(
                status="Published",
                binary_name="snapcraft",
                exact_match=True,
                distro_arch_series=distro_arch,
            ):
                yield distro, architecture, str(package.binary_package_version)


def request_autopkgtest_execution(cookie_path, distro, architecture, version):
    output = subprocess.check_output(
        [
            "wget",
            "-O-",
            "--load-cookies",
            cookie_path,
            "https://autopkgtest.ubuntu.com/request.cgi?release={distro}&"
            "arch={architecture}&package=snapcraft&"
            "ppa=snappy-dev/snapcraft-daily&"
            "trigger=snapcraft/{version}".format(
                distro=distro, architecture=architecture, version=version
            ),
        ]
    )
    if "Test request submitted" not in output:
        exit("Failed to request the autopkgtest")


if __name__ == "__main__":
    main()
